from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel, Field


class Collaborator(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    role: Optional[str] = None
    avatar: Optional[str] = None


class Commit(BaseModel):
    id: Optional[str] = None
    message: Optional[str] = None
    authorName: Optional[str] = None
    authorId: Optional[str] = None
    authorAvatar: Optional[str] = None
    branchName: Optional[str] = None
    createdAt: Optional[datetime] = None
    sourceApplication: Optional[str] = None
    referencedObject: Optional[str] = None
    totalChildrenCount: Optional[int] = None
    parents: Optional[List[str]] = None

    def __repr__(self) -> str:
        return (
            f"Commit( id: {self.id}, message: {self.message}, referencedObject:"
            f" {self.referencedObject}, authorName: {self.authorName}, branchName:"
            f" {self.branchName}, createdAt: {self.createdAt} )"
        )

    def __str__(self) -> str:
        return self.__repr__()


class Commits(BaseModel):
    totalCount: Optional[int] = None
    cursor: Optional[datetime] = None
    items: List[Commit] = []


class Object(BaseModel):
    id: Optional[str] = None
    speckleType: Optional[str] = None
    applicationId: Optional[str] = None
    totalChildrenCount: Optional[int] = None
    createdAt: Optional[datetime] = None


class Branch(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    commits: Optional[Commits] = None


class Branches(BaseModel):
    totalCount: Optional[int] = None
    cursor: Optional[datetime] = None
    items: List[Branch] = []


class Stream(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    role: Optional[str] = None
    isPublic: Optional[bool] = None
    description: Optional[str] = None
    createdAt: Optional[datetime] = None
    updatedAt: Optional[datetime] = None
    collaborators: List[Collaborator] = Field(default_factory=list)
    branches: Optional[Branches] = None
    commit: Optional[Commit] = None
    object: Optional[Object] = None
    commentCount: Optional[int] = None
    favoritedDate: Optional[datetime] = None
    favoritesCount: Optional[int] = None

    def __repr__(self):
        return (
            f"Stream( id: {self.id}, name: {self.name}, description:"
            f" {self.description}, isPublic: {self.isPublic})"
        )

    def __str__(self) -> str:
        return self.__repr__()


class Streams(BaseModel):
    totalCount: Optional[int] = None
    cursor: Optional[datetime] = None
    items: List[Stream] = []


class User(BaseModel):
    id: Optional[str] = None
    email: Optional[str] = None
    name: Optional[str] = None
    bio: Optional[str] = None
    company: Optional[str] = None
    avatar: Optional[str] = None
    verified: Optional[bool] = None
    role: Optional[str] = None
    streams: Optional[Streams] = None

    def __repr__(self):
        return (
            f"User( id: {self.id}, name: {self.name}, email: {self.email}, company:"
            f" {self.company} )"
        )

    def __str__(self) -> str:
        return self.__repr__()


class LimitedUser(BaseModel):
    """Limited user type, for showing public info about a user to another user."""

    id: str
    name: Optional[str] = None
    bio: Optional[str] = None
    company: Optional[str] = None
    avatar: Optional[str] = None
    verified: Optional[bool] = None
    role: Optional[str] = None


class PendingStreamCollaborator(BaseModel):
    id: Optional[str] = None
    inviteId: Optional[str] = None
    streamId: Optional[str] = None
    streamName: Optional[str] = None
    title: Optional[str] = None
    role: Optional[str] = None
    invitedBy: Optional[User] = None
    user: Optional[User] = None
    token: Optional[str] = None

    def __repr__(self):
        return (
            f"PendingStreamCollaborator( inviteId: {self.inviteId}, streamId:"
            f" {self.streamId}, role: {self.role}, title: {self.title}, invitedBy:"
            f" {self.user.name if self.user else None})"
        )

    def __str__(self) -> str:
        return self.__repr__()


class Activity(BaseModel):
    actionType: Optional[str] = None
    info: Optional[dict] = None
    userId: Optional[str] = None
    streamId: Optional[str] = None
    resourceId: Optional[str] = None
    resourceType: Optional[str] = None
    message: Optional[str] = None
    time: Optional[datetime] = None

    def __repr__(self) -> str:
        return (
            f"Activity( streamId: {self.streamId}, actionType: {self.actionType},"
            f" message: {self.message}, userId: {self.userId} )"
        )

    def __str__(self) -> str:
        return self.__repr__()


class ActivityCollection(BaseModel):
    totalCount: Optional[int] = None
    items: Optional[List[Activity]] = None
    cursor: Optional[datetime] = None

    def __repr__(self) -> str:
        return (
            f"ActivityCollection( totalCount: {self.totalCount}, items:"
            f" {len(self.items) if self.items else 0}, cursor:"
            f" {self.cursor.isoformat() if self.cursor else None} )"
        )

    def __str__(self) -> str:
        return self.__repr__()


class ServerInfo(BaseModel):
    name: Optional[str] = None
    company: Optional[str] = None
    url: Optional[str] = None
    description: Optional[str] = None
    adminContact: Optional[str] = None
    canonicalUrl: Optional[str] = None
    roles: Optional[List[dict]] = None
    scopes: Optional[List[dict]] = None
    authStrategies: Optional[List[dict]] = None
    version: Optional[str] = None



from typing import Any, List

from specklepy.objects.base import Base

NAMESPACE = "Objects.Primitive"


class Interval(Base, speckle_type=f"{NAMESPACE}.Interval"):
    start: float = 0.0
    end: float = 0.0

    def length(self):
        return abs(self.start - self.end)

    @classmethod
    def from_list(cls, args: List[Any]) -> "Interval":
        return cls(start=args[0], end=args[1])

    def to_list(self) -> List[Any]:
        return [self.start, self.end]


class Interval2d(Base, speckle_type=f"{NAMESPACE}.Interval2d"):
    u: Interval
    v: Interval



from datetime import datetime, timezone
from typing import List, Optional

from gql import gql

from specklepy.api.models import ActivityCollection, PendingStreamCollaborator, User
from specklepy.api.resource import ResourceBase
from specklepy.logging import metrics
from specklepy.logging.exceptions import SpeckleException

from specklepy.core.api.resources.active_user import Resource as CoreResource


class Resource(CoreResource):
    """API Access class for users"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            server_version=server_version,
        )
        self.schema = User

    def get(self) -> User:
        """Gets the profile of a user. If no id argument is provided,
        will return the current authenticated user's profile
        (as extracted from the authorization header).

        Arguments:
            id {str} -- the user id

        Returns:
            User -- the retrieved user
        """
        metrics.track(metrics.SDK, custom_props={"name": "User Active Get"})
        return super().get()

    def update(
        self,
        name: Optional[str] = None,
        company: Optional[str] = None,
        bio: Optional[str] = None,
        avatar: Optional[str] = None,
    ):
        """Updates your user profile. All arguments are optional.

        Arguments:
            name {str} -- your name
            company {str} -- the company you may or may not work for
            bio {str} -- tell us about yourself
            avatar {str} -- a nice photo of yourself

        Returns    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT):
            bool -- True if your profile was updated successfully
        """
        metrics.track(metrics.SDK, self.account, {"name": "User Active Update"})
        return super().update(name, company, bio, avatar)

    def activity(
        self,
        limit: int = 20,
        action_type: Optional[str] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        cursor: Optional[datetime] = None,
    ):
        """
        Get the activity from a given stream in an Activity collection.
        Step into the activity `items` for the list of activity.
        If no id argument is provided, will return the current authenticated user's
        activity (as extracted from the authorization header).

        Note: all timestamps arguments should be `datetime` of any tz as they will be
        converted to UTC ISO format strings

        user_id {str} -- the id of the user to get the activity from
        action_type {str} -- filter results to a single action type
            (eg: `commit_create` or `commit_receive`)
        limit {int} -- max number of Activity items to return
        before {datetime} -- latest cutoff for activity
            (ie: return all activity _before_ this time)
        after {datetime} -- oldest cutoff for activity
            (ie: return all activity _after_ this time)
        cursor {datetime} -- timestamp cursor for pagination
        """
        metrics.track(metrics.SDK, self.account, {"name": "User Active Activity"})
        return super().activity(limit, action_type, before, after, cursor)

    def get_all_pending_invites(self) -> List[PendingStreamCollaborator]:
        """Get all of the active user's pending stream invites

        Requires Speckle Server version >= 2.6.4

        Returns:
            List[PendingStreamCollaborator]
            -- a list of pending invites for the current user
        """
        metrics.track(metrics.SDK, self.account, {"name": "User Active Invites All Get"})
        return super().get_all_pending_invites()

    def get_pending_invite(
        self, stream_id: str, token: Optional[str] = None
    ) -> Optional[PendingStreamCollaborator]:
        """Get a particular pending invite for the active user on a given stream.
        If no invite_id is provided, any valid invite will be returned.

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to look for invites on
            token {str} -- the token of the invite to look for (optional)

        Returns:
            PendingStreamCollaborator
            -- the invite for the given stream (or None if it isn't found)
        """
        metrics.track(metrics.SDK, self.account, {"name": "User Active Invite Get"})
        return super().get_pending_invite(stream_id, token)



"""Builtin Speckle object kit."""

from specklepy.objects import encoding, geometry, other, primitive, structural, units
from specklepy.objects.base import Base

__all__ = ["Base", "encoding", "geometry", "other", "units", "structural", "primitive"]




from typing import Optional, Union, List 
from specklepy.objects.geometry import Point, Line, Polyline, Circle, Arc, Polycurve, Mesh 
from specklepy.objects import Base
from deprecated import deprecated

class GisPolygonGeometry(Base, speckle_type="Objects.GIS.PolygonGeometry", detachable={"displayValue"}):
    """GIS Polygon Geometry"""

    boundary: Optional[Union[Polyline, Arc, Line, Circle, Polycurve]] = None
    voids: Optional[List[Union[Polyline, Arc, Line, Circle, Polycurve]] ] = None
    displayValue: Optional[List[Mesh]] = None

class GisPolygonElement(Base, speckle_type="Objects.GIS.PolygonElement"):
    """GIS Polygon element"""

    geometry: Optional[List[GisPolygonGeometry]] = None
    attributes: Optional[Base] = None

class GisLineElement(Base, speckle_type="Objects.GIS.LineElement"):
    """GIS Polyline element"""

    geometry: Optional[List[Union[Polyline, Arc, Line, Circle, Polycurve]]] = None,
    attributes: Optional[Base] = None,

class GisPointElement(Base, speckle_type="Objects.GIS.PointElement"):
    """GIS Point element"""

    geometry: Optional[List[Point]] = None,
    attributes: Optional[Base] = None,

class GisRasterElement(Base, speckle_type="Objects.GIS.RasterElement", detachable={"displayValue"}):
    """GIS Raster element"""

    band_count: Optional[int] = None
    band_names: Optional[List[str]] = None
    x_origin: Optional[float] = None
    y_origin: Optional[float] = None
    x_size: Optional[int] = None
    y_size: Optional[int] = None
    x_resolution: Optional[float] = None
    y_resolution: Optional[float] = None
    noDataValue: Optional[List[float]] = None
    displayValue: Optional[List[Mesh]] = None

class GisTopography(GisRasterElement, speckle_type="Objects.GIS.GisTopography", detachable={"displayValue"}):
    """GIS Raster element with 3d Topography representation"""

class GisNonGeometryElement(Base, speckle_type="Objects.GIS.NonGeometryElement"):
    """GIS Table feature"""

    attributes: Optional[Base] = None




"""This module provides an abstraction layer above the Speckle Automate runtime."""
from dataclasses import dataclass, field
from pathlib import Path
import time
from typing import Any, Dict, List, Optional, Tuple, Union

import httpx
from gql import gql
from specklepy.api import operations
from specklepy.api.client import SpeckleClient
from specklepy.core.api.models import Branch
from specklepy.objects import Base
from specklepy.transports.memory import MemoryTransport
from specklepy.transports.server import ServerTransport
from specklepy.logging.exceptions import SpeckleException

from speckle_automate.schema import (
    AutomateBase,
    AutomationResult,
    AutomationRunData,
    AutomationStatus,
    ObjectResultLevel,
    ResultCase,
)


@dataclass
class AutomationContext:
    """A context helper class.

    This class exposes methods to work with the Speckle Automate context inside
    Speckle Automate functions.

    An instance of AutomationContext is injected into every run of a function.
    """

    automation_run_data: AutomationRunData
    speckle_client: SpeckleClient
    _server_transport: ServerTransport
    _speckle_token: str

    #: keep a memory transponrt at hand, to speed up things if needed
    _memory_transport: MemoryTransport = field(default_factory=MemoryTransport)

    #: added for performance measuring
    _init_time: float = field(default_factory=time.perf_counter)
    _automation_result: AutomationResult = field(default_factory=AutomationResult)

    @classmethod
    def initialize(
        cls, automation_run_data: Union[str, AutomationRunData], speckle_token: str
    ) -> "AutomationContext":
        """Bootstrap the AutomateSDK from raw data.

        Todo:
        ----
            * bootstrap a structlog logger instance
            * expose a logger, that ppl can use instead of print
        """
        # parse the json value if its not an initialized project data instance
        automation_run_data = (
            automation_run_data
            if isinstance(automation_run_data, AutomationRunData)
            else AutomationRunData.model_validate_json(automation_run_data)
        )
        speckle_client = SpeckleClient(
            automation_run_data.speckle_server_url,
            automation_run_data.speckle_server_url.startswith("https"),
        )
        speckle_client.authenticate_with_token(speckle_token)
        if not speckle_client.account:
            msg = (
                f"Could not autenticate to {automation_run_data.speckle_server_url}",
                "with the provided token",
            )
            raise ValueError(msg)
        server_transport = ServerTransport(
            automation_run_data.project_id, speckle_client
        )
        return cls(automation_run_data, speckle_client, server_transport, speckle_token)

    @property
    def run_status(self) -> AutomationStatus:
        """Get the status of the automation run."""
        return self._automation_result.run_status

    @property
    def status_message(self) -> Optional[str]:
        """Get the current status message."""
        return self._automation_result.status_message

    def elapsed(self) -> float:
        """Return the elapsed time in seconds since the initialization time."""
        return time.perf_counter() - self._init_time

    def receive_version(self) -> Base:
        """Receive the Speckle project version that triggered this automation run."""
        commit = self.speckle_client.commit.get(
            self.automation_run_data.project_id, self.automation_run_data.version_id
        )
        if not commit.referencedObject:
            raise ValueError("The commit has no referencedObject, cannot receive it.")
        base = operations.receive(
            commit.referencedObject, self._server_transport, self._memory_transport
        )
        print(
            f"It took {self.elapsed():.2f} seconds to receive",
            f" the speckle version {self.automation_run_data.version_id}",
        )
        return base

    def create_new_version_in_project(
        self, root_object: Base, model_name: str, version_message: str = ""
    ) -> Tuple[str, str]:
        """Save a base model to a new version on the project.

        Args:
            root_object (Base): The Speckle base object for the new version.
            model_id (str): For now please use a `branchName`!
            version_message (str): The message for the new version.
        """

        if model_name == self.automation_run_data.branch_name:
            raise ValueError(
                f"The target model: {model_name} cannot match the model"
                f" that triggered this automation:"
                f" {self.automation_run_data.model_id} /"
                f" {self.automation_run_data.branch_name}"
            )

        branch = self.speckle_client.branch.get(
            self.automation_run_data.project_id, model_name, 1
        )
        # we just check if it exists
        if (not branch) or isinstance(branch, SpeckleException):
            branch_create = self.speckle_client.branch.create(
                self.automation_run_data.project_id,
                model_name,
            )
            if isinstance(branch_create, Exception):
                raise branch_create
            model_id = branch_create
        else:
            model_id = branch.id

        root_object_id = operations.send(
            root_object,
            [self._server_transport, self._memory_transport],
            use_default_cache=False,
        )

        version_id = self.speckle_client.commit.create(
            stream_id=self.automation_run_data.project_id,
            object_id=root_object_id,
            branch_name=model_name,
            message=version_message,
            source_application="SpeckleAutomate",
        )

        if isinstance(version_id, SpeckleException):
            raise version_id

        self._automation_result.result_versions.append(version_id)
        return model_id, version_id

    def set_context_view(
        self,
        # f"{model_id}@{version_id} or {model_id} "
        resource_ids: Optional[List[str]] = None,
        include_source_model_version: bool = True,
    ) -> None:
        link_resources = (
            [
                f"{self.automation_run_data.model_id}@{self.automation_run_data.version_id}"
            ]
            if include_source_model_version
            else []
        )
        if resource_ids:
            link_resources.append(*resource_ids)
        if not link_resources:
            raise Exception(
                "We do not have enough resource ids to compose a context view"
            )
        self._automation_result.result_view = (
            f"/projects/{self.automation_run_data.project_id}"
            f"/models/{','.join(link_resources)}"
        )

    def report_run_status(self) -> None:
        """Report the current run status to the project of this automation."""
        query = gql(
            """
            mutation ReportFunctionRunStatus(
                $automationId: String!, 
                $automationRevisionId: String!, 
                $automationRunId: String!,
                $versionId: String!,
                $functionId: String!,
                $functionName: String!,
                $functionLogo: String,
                $runStatus: AutomationRunStatus!
                $elapsed: Float!
                $contextView: String
                $resultVersionIds: [String!]!
                $statusMessage: String
                $objectResults: JSONObject
            ){
                automationMutations {
                    functionRunStatusReport(input: {
                        automationId: $automationId
                        automationRevisionId: $automationRevisionId
                        automationRunId: $automationRunId
                        versionId: $versionId
                        functionRuns: [
                        {
                            functionId: $functionId
                            functionName: $functionName
                            functionLogo: $functionLogo
                            status: $runStatus,
                            contextView: $contextView,
                            elapsed: $elapsed,
                            resultVersionIds: $resultVersionIds,
                            statusMessage: $statusMessage
                            results: $objectResults
                        }]
                   })
                }
            }
            """
        )
        if self.run_status in [AutomationStatus.SUCCEEDED, AutomationStatus.FAILED]:
            object_results = {
                "version": "1.0.0",
                "values": {
                    "objectResults": self._automation_result.model_dump(by_alias=True)[
                        "objectResults"
                    ],
                    "blobIds": self._automation_result.blobs,
                },
            }
        else:
            object_results = None

        params = {
            "automationId": self.automation_run_data.automation_id,
            "automationRevisionId": self.automation_run_data.automation_revision_id,
            "automationRunId": self.automation_run_data.automation_run_id,
            "versionId": self.automation_run_data.version_id,
            "functionId": self.automation_run_data.function_id,
            "functionName": self.automation_run_data.function_name,
            "functionLogo": self.automation_run_data.function_logo,
            "runStatus": self.run_status.value,
            "statusMessage": self._automation_result.status_message,
            "contextView": self._automation_result.result_view,
            "elapsed": self.elapsed(),
            "resultVersionIds": self._automation_result.result_versions,
            "objectResults": object_results,
        }
        self.speckle_client.httpclient.execute(query, params)

    def store_file_result(self, file_path: Union[Path, str]) -> None:
        """Save a file attached to the project of this automation."""
        path_obj = (
            Path(file_path).resolve() if isinstance(file_path, str) else file_path
        )

        class UploadResult(AutomateBase):
            blob_id: str
            file_name: str
            upload_status: int

        class BlobUploadResponse(AutomateBase):
            upload_results: list[UploadResult]

        if not path_obj.exists():
            raise ValueError("The given file path doesn't exist")
        files = {path_obj.name: open(str(path_obj), "rb")}

        url = (
            f"{self.automation_run_data.speckle_server_url}/api/stream/"
            f"{self.automation_run_data.project_id}/blob"
        )
        data = (
            httpx.post(
                url,
                files=files,
                headers={"authorization": f"Bearer {self._speckle_token}"},
            )
            .raise_for_status()
            .json()
        )

        upload_response = BlobUploadResponse.model_validate(data)

        if len(upload_response.upload_results) != 1:
            raise ValueError("Expecting one upload result.")

        self._automation_result.blobs.extend(
            [upload_result.blob_id for upload_result in upload_response.upload_results]
        )

    def mark_run_failed(self, status_message: str) -> None:
        """Mark the current run a failure."""
        self._mark_run(AutomationStatus.FAILED, status_message)

    def mark_run_success(self, status_message: Optional[str]) -> None:
        """Mark the current run a success with an optional message."""
        self._mark_run(AutomationStatus.SUCCEEDED, status_message)

    def _mark_run(
        self, status: AutomationStatus, status_message: Optional[str]
    ) -> None:
        duration = self.elapsed()
        self._automation_result.status_message = status_message
        self._automation_result.run_status = status
        self._automation_result.elapsed = duration

        msg = f"Automation run {status.value} after {duration:.2f} seconds."
        print("\n".join([msg, status_message]) if status_message else msg)

    def attach_error_to_objects(
        self,
        category: str,
        object_ids: Union[str, List[str]],
        message: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        visual_overrides: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a new error case to the run results.

        If the error cause has already created an error case,
        the error will be extended with a new case refering to the causing objects.
        Args:
            error_tag (str): A short tag for the error type.
            causing_object_ids (str[]): A list of object_id-s that are causing the error
            error_messagge (Optional[str]): Optional error message.
            metadata: User provided metadata key value pairs
            visual_overrides: Case specific 3D visual overrides.
        """
        self.attach_result_to_objects(
            ObjectResultLevel.ERROR,
            category,
            object_ids,
            message,
            metadata,
            visual_overrides,
        )

    def attach_warning_to_objects(
        self,
        category: str,
        object_ids: Union[str, List[str]],
        message: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        visual_overrides: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a new warning case to the run results."""
        self.attach_result_to_objects(
            ObjectResultLevel.WARNING,
            category,
            object_ids,
            message,
            metadata,
            visual_overrides,
        )

    def attach_info_to_objects(
        self,
        category: str,
        object_ids: Union[str, List[str]],
        message: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        visual_overrides: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a new info case to the run results."""
        self.attach_result_to_objects(
            ObjectResultLevel.INFO,
            category,
            object_ids,
            message,
            metadata,
            visual_overrides,
        )

    def attach_result_to_objects(
        self,
        level: ObjectResultLevel,
        category: str,
        object_ids: Union[str, List[str]],
        message: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        visual_overrides: Optional[Dict[str, Any]] = None,
    ) -> None:
        if isinstance(object_ids, list):
            if len(object_ids) < 1:
                raise ValueError(
                    f"Need atleast one object_id to report a(n) {level.value.upper()}"
                )
            id_list = object_ids
        else:
            id_list = [object_ids]
        print(
            f"Object {', '.join(id_list)} was marked with {level.value.upper()}",
            f"/{category} cause: {message}",
        )
        self._automation_result.object_results.append(
            ResultCase(
                category=category,
                level=level,
                object_ids=id_list,
                message=message,
                metadata=metadata,
                visual_overrides=visual_overrides,
            )
        )



"""
This is the Core SDK module of `specklepy`.

This module should be kept in sync with the functionalities of our other SDKs especially
C# Core https://github.com/specklesystems/speckle-sharp/tree/main/Core/Core
"""



from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel, Field

# following imports seem to be unnecessary, but they need to stay 
# to not break the scripts using these functions as non-core
from specklepy.core.api.models import (Collaborator, Commit,
                                       Commits, Object, Branch, Branches,
                                       Stream, Streams, User, LimitedUser,
                                       PendingStreamCollaborator, Activity,
                                       ActivityCollection, ServerInfo)



from dataclasses import dataclass
from enum import Enum
from unicodedata import name

# following imports seem to be unnecessary, but they need to stay 
# to not break the scripts using these functions as non-core
from specklepy.core.api.host_applications import (HostApplication, HostAppVersion,
                                                  get_host_app_from_string, 
                                                  _app_name_host_app_mapping,
                                                  RHINO,GRASSHOPPER,REVIT,DYNAMO,UNITY,GSA,
                                                  CIVIL,AUTOCAD,MICROSTATION,OPENROADS,
                                                  OPENRAIL,OPENBUILDINGS,ETABS,SAP2000,CSIBRIDGE,
                                                  SAFE,TEKLASTRUCTURES,DXF,EXCEL,UNREAL,POWERBI,
                                                  BLENDER,QGIS,ARCGIS,SKETCHUP,ARCHICAD,TOPSOLID,
                                                  PYTHON,NET,OTHER)

if __name__ == "__main__":
    print(HostAppVersion.v)



from datetime import datetime, timezone
from typing import List, Optional

from deprecated import deprecated
from gql import gql

from specklepy.core.api.models import ActivityCollection, PendingStreamCollaborator, Stream
from specklepy.core.api.resource import ResourceBase
from specklepy.logging.exceptions import SpeckleException, UnsupportedException

NAME = "stream"


class Resource(ResourceBase):
    """API Access class for streams"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
            server_version=server_version,
        )

        self.schema = Stream

    def get(self, id: str, branch_limit: int = 10, commit_limit: int = 10) -> Stream:
        """Get the specified stream from the server

        Arguments:
            id {str} -- the stream id
            branch_limit {int} -- the maximum number of branches to return
            commit_limit {int} -- the maximum number of commits to return

        Returns:
            Stream -- the retrieved stream
        """
        query = gql(
            """
            query Stream($id: String!, $branch_limit: Int!, $commit_limit: Int!) {
                stream(id: $id) {
                    id
                    name
                    role
                    description
                    isPublic
                    createdAt
                    updatedAt
                    commentCount
                    favoritesCount
                    collaborators {
                        id
                        name
                        role
                        avatar
                    }
                    branches(limit: $branch_limit) {
                        totalCount
                        cursor
                        items {
                            id
                            name
                            description
                            commits(limit: $commit_limit) {
                                totalCount
                                cursor
                                items {
                                    id
                                    message
                                    authorId
                                    createdAt
                                    authorName
                                    referencedObject
                                    sourceApplication
                                }
                              }
                          }
                      }
                }
            }
          """
        )

        params = {"id": id, "branch_limit": branch_limit, "commit_limit": commit_limit}

        return self.make_request(query=query, params=params, return_type="stream")

    def list(self, stream_limit: int = 10) -> List[Stream]:
        """Get a list of the user's streams

        Arguments:
            stream_limit {int} -- The maximum number of streams to return

        Returns:
            List[Stream] -- A list of Stream objects
        """
        query = gql(
            """
            query User($stream_limit: Int!) {
                user {
                    id
                    bio
                    name
                    email
                    avatar
                    company
                    verified
                    profiles
                    role
                    streams(limit: $stream_limit) {
                        totalCount
                        cursor
                        items {
                            id
                            name
                            role
                            isPublic
                            createdAt
                            updatedAt
                            description
                            commentCount
                            favoritesCount
                            collaborators {
                                id
                                name
                                role
                            }
                          }
                      }
                  }
            }
          """
        )

        params = {"stream_limit": stream_limit}

        return self.make_request(
            query=query, params=params, return_type=["user", "streams", "items"]
        )

    def create(
        self,
        name: str = "Anonymous Python Stream",
        description: str = "No description provided",
        is_public: bool = True,
    ) -> str:
        """Create a new stream

        Arguments:
            name {str} -- the name of the string
            description {str} -- a short description of the stream
            is_public {bool}
                -- whether or not the stream can be viewed by anyone with the id

        Returns:
            id {str} -- the id of the newly created stream
        """
        query = gql(
            """
            mutation StreamCreate($stream: StreamCreateInput!) {
              streamCreate(stream: $stream)
            }
        """
        )

        params = {
            "stream": {"name": name, "description": description, "isPublic": is_public}
        }

        return self.make_request(
            query=query, params=params, return_type="streamCreate", parse_response=False
        )

    def update(
        self,
        id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_public: Optional[bool] = None,
    ) -> bool:
        """Update an existing stream

        Arguments:
            id {str} -- the id of the stream to be updated
            name {str} -- the name of the string
            description {str} -- a short description of the stream
            is_public {bool}
                -- whether or not the stream can be viewed by anyone with the id

        Returns:
            bool -- whether the stream update was successful
        """
        query = gql(
            """
            mutation StreamUpdate($stream: StreamUpdateInput!) {
                streamUpdate(stream: $stream)
            }
        """
        )

        params = {
            "id": id,
            "name": name,
            "description": description,
            "isPublic": is_public,
        }
        # remove None values so graphql doesn't cry
        params = {"stream": {k: v for k, v in params.items() if v is not None}}

        return self.make_request(
            query=query, params=params, return_type="streamUpdate", parse_response=False
        )

    def delete(self, id: str) -> bool:
        """Delete a stream given its id

        Arguments:
            id {str} -- the id of the stream to delete

        Returns:
            bool -- whether the deletion was successful
        """
        query = gql(
            """
            mutation StreamDelete($id: String!) {
                streamDelete(id: $id)
            }
            """
        )

        params = {"id": id}

        return self.make_request(
            query=query, params=params, return_type="streamDelete", parse_response=False
        )

    def search(
        self,
        search_query: str,
        limit: int = 25,
        branch_limit: int = 10,
        commit_limit: int = 10,
    ):
        """Search for streams by name, description, or id

        Arguments:
            search_query {str} -- a string to search for
            limit {int} -- the maximum number of results to return
            branch_limit {int} -- the maximum number of branches to return
            commit_limit {int} -- the maximum number of commits to return

        Returns:
            List[Stream] -- a list of Streams that match the search query
        """
        query = gql(
            """
            query StreamSearch(
                $search_query: String!,
                $limit: Int!,
                $branch_limit:Int!,
                $commit_limit:Int!
            ) {
                streams(query: $search_query, limit: $limit) {
                    items {
                        id
                        name
                        role
                        description
                        isPublic
                        createdAt
                        updatedAt
                        collaborators {
                            id
                            name
                            role
                            avatar
                        }
                        branches(limit: $branch_limit) {
                            totalCount
                            cursor
                            items {
                                id
                                name
                                description
                                commits(limit: $commit_limit) {
                                    totalCount
                                    cursor
                                    items {
                                        id
                                        referencedObject
                                        message
                                        authorName
                                        authorId
                                        createdAt
                                    }
                                }
                            }
                        }
                    }
                }
            }
          """
        )

        params = {
            "search_query": search_query,
            "limit": limit,
            "branch_limit": branch_limit,
            "commit_limit": commit_limit,
        }

        return self.make_request(
            query=query, params=params, return_type=["streams", "items"]
        )

    def favorite(self, stream_id: str, favorited: bool = True):
        """Favorite or unfavorite the given stream.

        Arguments:
            stream_id {str} -- the id of the stream to favorite / unfavorite
            favorited {bool}
                -- whether to favorite (True) or unfavorite (False) the stream

        Returns:
            Stream -- the stream with its `id`, `name`, and `favoritedDate`
        """
        query = gql(
            """
            mutation StreamFavorite($stream_id: String!, $favorited: Boolean!) {
                streamFavorite(streamId: $stream_id, favorited: $favorited) {
                    id
                    name
                    favoritedDate
                    favoritesCount
                }
            }
            """
        )

        params = {
            "stream_id": stream_id,
            "favorited": favorited,
        }

        return self.make_request(
            query=query, params=params, return_type=["streamFavorite"]
        )

    def get_all_pending_invites(
        self, stream_id: str
    ) -> List[PendingStreamCollaborator]:
        """Get all of the pending invites on a stream.
        You must be a `stream:owner` to query this.

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the stream id from which to get the pending invites

        Returns:
            List[PendingStreamCollaborator]
                -- a list of pending invites for the specified stream
        """
        self._check_invites_supported()

        query = gql(
            """
            query StreamInvites($streamId: String!) {
                stream(id: $streamId){
                    pendingCollaborators {
                        id
                        token
                        inviteId
                        streamId
                        streamName
                        title
                        role
                        invitedBy{
                            id
                            name
                            company
                            avatar
                        }
                        user {
                            id
                            name
                            company
                            avatar
                        }
                    }
                }
            }
            """
        )
        params = {"streamId": stream_id}

        return self.make_request(
            query=query,
            params=params,
            return_type=["stream", "pendingCollaborators"],
            schema=PendingStreamCollaborator,
        )

    def invite(
        self,
        stream_id: str,
        email: Optional[str] = None,
        user_id: Optional[str] = None,
        role: str = "stream:contributor",  # should default be reviewer?
        message: Optional[str] = None,
    ):
        """Invite someone to a stream using either their email or user id

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to invite the user to
            email {str} -- the email of the user to invite (use this OR `user_id`)
            user_id {str} -- the id of the user to invite (use this OR `email`)
            role {str}
                -- the role to assign to the user (defaults to `stream:contributor`)
            message {str}
                -- a message to send along with this invite to the specified user

        Returns:
            bool -- True if the operation was successful
        """
        self._check_invites_supported()

        if email is None and user_id is None:
            raise SpeckleException(
                "You must provide either an email or a user id to use the"
                " `stream.invite` method"
            )

        query = gql(
            """
            mutation StreamInviteCreate($input: StreamInviteCreateInput!) {
                streamInviteCreate(input: $input)
            }
            """
        )

        params = {
            "email": email,
            "userId": user_id,
            "streamId": stream_id,
            "message": message,
            "role": role,
        }
        params = {"input": {k: v for k, v in params.items() if v is not None}}

        return self.make_request(
            query=query,
            params=params,
            return_type="streamInviteCreate",
            parse_response=False,
        )

    def invite_batch(
        self,
        stream_id: str,
        emails: Optional[List[str]] = None,
        user_ids: Optional[List[None]] = None,
        message: Optional[str] = None,
    ) -> bool:
        """Invite a batch of users to a specified stream.

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to invite the user to
            emails {List[str]}
                -- the email of the user to invite (use this and/or `user_ids`)
            user_id {List[str]}
                -- the id of the user to invite (use this and/or `emails`)
            message {str}
                -- a message to send along with this invite to the specified user

        Returns:
            bool -- True if the operation was successful
        """
        self._check_invites_supported()
        if emails is None and user_ids is None:
            raise SpeckleException(
                "You must provide either an email or a user id to use the"
                " `stream.invite` method"
            )

        query = gql(
            """
            mutation StreamInviteBatchCreate($input: [StreamInviteCreateInput!]!) {
                streamInviteBatchCreate(input: $input)
            }
            """
        )

        email_invites = [
            {"streamId": stream_id, "message": message, "email": email}
            for email in (emails if emails is not None else [])
            if email is not None
        ]

        user_invites = [
            {"streamId": stream_id, "message": message, "userId": user_id}
            for user_id in (user_ids if user_ids is not None else []) 
            if user_id is not None
        ]


        params = {"input": [*email_invites, *user_invites]}

        return self.make_request(
            query=query,
            params=params,
            return_type="streamInviteBatchCreate",
            parse_response=False,
        )

    def invite_cancel(self, stream_id: str, invite_id: str) -> bool:
        """Cancel an existing stream invite

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream invite
            invite_id {str} -- the id of the invite to use

        Returns:
            bool -- true if the operation was successful
        """
        self._check_invites_supported()

        query = gql(
            """
            mutation StreamInviteCancel($streamId: String!, $inviteId: String!) {
                streamInviteCancel(streamId: $streamId, inviteId: $inviteId)
            }
            """
        )

        params = {"streamId": stream_id, "inviteId": invite_id}

        return self.make_request(
            query=query,
            params=params,
            return_type="streamInviteCancel",
            parse_response=False,
        )

    def invite_use(self, stream_id: str, token: str, accept: bool = True) -> bool:
        """Accept or decline a stream invite

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str}
                -- the id of the stream for which the user has a pending invite
            token {str} -- the token of the invite to use
            accept {bool} -- whether or not to accept the invite (defaults to True)

        Returns:
            bool -- true if the operation was successful
        """
        self._check_invites_supported()

        query = gql(
            """
            mutation StreamInviteUse(
                $accept: Boolean!,
                $streamId: String!,
                $token: String!
                ) {
                streamInviteUse(accept: $accept, streamId: $streamId, token: $token)
            }
            """
        )

        params = {"streamId": stream_id, "token": token, "accept": accept}

        return self.make_request(
            query=query,
            params=params,
            return_type="streamInviteUse",
            parse_response=False,
        )

    def update_permission(self, stream_id: str, user_id: str, role: str):
        """Updates permissions for a user on a given stream

        Valid for Speckle Server >=2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to grant permissions to
            user_id {str} -- the id of the user to grant permissions for
            role {str} -- the role to grant the user

        Returns:
            bool -- True if the operation was successful
        """
        if self.server_version and (
            self.server_version != ("dev",) and self.server_version < (2, 6, 4)
        ):
            raise UnsupportedException(
                "Server mutation `update_permission` is only supported as of Speckle"
                " Server v2.6.4. Please update your Speckle Server to use this method"
                " or use the `grant_permission` method instead."
            )
        query = gql(
            """
            mutation StreamUpdatePermission(
                $permission_params: StreamUpdatePermissionInput!
                ) {
                streamUpdatePermission(permissionParams: $permission_params)
            }
            """
        )

        params = {
            "permission_params": {
                "streamId": stream_id,
                "userId": user_id,
                "role": role,
            }
        }

        return self.make_request(
            query=query,
            params=params,
            return_type="streamUpdatePermission",
            parse_response=False,
        )

    def revoke_permission(self, stream_id: str, user_id: str):
        """Revoke permissions from a user on a given stream

        Arguments:
            stream_id {str} -- the id of the stream to revoke permissions from
            user_id {str} -- the id of the user to revoke permissions from

        Returns:
            bool -- True if the operation was successful
        """
        query = gql(
            """
            mutation StreamRevokePermission(
                $permission_params: StreamRevokePermissionInput!
                ) {
                streamRevokePermission(permissionParams: $permission_params)
            }
            """
        )

        params = {"permission_params": {"streamId": stream_id, "userId": user_id}}

        return self.make_request(
            query=query,
            params=params,
            return_type="streamRevokePermission",
            parse_response=False,
        )

    def activity(
        self,
        stream_id: str,
        action_type: Optional[str] = None,
        limit: int = 20,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        cursor: Optional[datetime] = None,
    ):
        """
        Get the activity from a given stream in an Activity collection.
        Step into the activity `items` for the list of activity.

        Note: all timestamps arguments should be `datetime` of any tz
            as they will be converted to UTC ISO format strings

        stream_id {str} -- the id of the stream to get activity from
        action_type {str}
            -- filter results to a single action type
            (eg: `commit_create` or `commit_receive`)
        limit {int} -- max number of Activity items to return
        before {datetime}
            -- latest cutoff for activity (ie: return all activity _before_ this time)
        after {datetime}
            -- oldest cutoff for activity (ie: return all activity _after_ this time)
        cursor {datetime} -- timestamp cursor for pagination
        """
        query = gql(
            """
            query StreamActivity(
                $stream_id: String!,
                $action_type: String,
                $before:DateTime,
                $after: DateTime,
                $cursor: DateTime,
                $limit: Int
                ){
                stream(id: $stream_id) {
                    activity(
                        actionType: $action_type,
                        before: $before,
                        after: $after,
                        cursor: $cursor,
                        limit: $limit
                        ) {
                        totalCount
                        cursor
                        items {
                            actionType
                            info
                            userId
                            streamId
                            resourceId
                            resourceType
                            message
                            time
                        }
                    }
                }
            }
            """
        )
        try:
            params = {
                "stream_id": stream_id,
                "limit": limit,
                "action_type": action_type,
                "before": before.astimezone(timezone.utc).isoformat()
                if before
                else before,
                "after": after.astimezone(timezone.utc).isoformat() if after else after,
                "cursor": cursor.astimezone(timezone.utc).isoformat()
                if cursor
                else cursor,
            }
        except AttributeError as e:
            raise SpeckleException(
                "Could not get stream activity - `before`, `after`, and `cursor` must"
                " be in `datetime` format if provided",
                ValueError(),
            ) from e

        return self.make_request(
            query=query,
            params=params,
            return_type=["stream", "activity"],
            schema=ActivityCollection,
        )



from datetime import datetime, timezone
from typing import List, Optional, Union

from deprecated import deprecated
from gql import gql

from specklepy.core.api.models import ActivityCollection, PendingStreamCollaborator, User
from specklepy.core.api.resource import ResourceBase

from specklepy.logging.exceptions import SpeckleException

NAME = "user"

DEPRECATION_VERSION = "2.9.0"
DEPRECATION_TEXT = (
    "The user resource is deprecated, please use the active_user or other_user"
    " resources"
)


class Resource(ResourceBase):
    """API Access class for users"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
            server_version=server_version,
        )
        self.schema = User

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def get(self, id: Optional[str] = None) -> User:
        """
        Gets the profile of a user.
        If no id argument is provided, will return the current authenticated
        user's profile (as extracted from the authorization header).

        Arguments:
            id {str} -- the user id

        Returns:
            User -- the retrieved user
        """
        query = gql(
            """
            query User($id: String) {
                user(id: $id) {
                    id
                    email
                    name
                    bio
                    company
                    avatar
                    verified
                    profiles
                    role
                }
            }
          """
        )

        params = {"id": id}

        return self.make_request(query=query, params=params, return_type="user")

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def search(
        self, search_query: str, limit: int = 25
    ) -> Union[List[User], SpeckleException]:
        """
        Searches for user by name or email.
        The search query must be at least 3 characters long

        Arguments:
            search_query {str} -- a string to search for
            limit {int} -- the maximum number of results to return
        Returns:
            List[User] -- a list of User objects that match the search query
        """
        if len(search_query) < 3:
            return SpeckleException(
                message="User search query must be at least 3 characters"
            )

        query = gql(
            """
            query UserSearch($search_query: String!, $limit: Int!) {
                userSearch(query: $search_query, limit: $limit) {
                    items {
                        id
                        name
                        bio
                        company
                        avatar
                        verified
                    }
                }
            }
          """
        )
        params = {"search_query": search_query, "limit": limit}

        return self.make_request(
            query=query, params=params, return_type=["userSearch", "items"]
        )

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def update(
        self,
        name: Optional[str] = None,
        company: Optional[str] = None,
        bio: Optional[str] = None,
        avatar: Optional[str] = None,
    ):
        """Updates your user profile. All arguments are optional.

        Arguments:
            name {str} -- your name
            company {str} -- the company you may or may not work for
            bio {str} -- tell us about yourself
            avatar {str} -- a nice photo of yourself

        Returns:
            bool -- True if your profile was updated successfully
        """
        query = gql(
            """
            mutation UserUpdate($user: UserUpdateInput!) {
                userUpdate(user: $user)
            }
            """
        )
        params = {"name": name, "company": company, "bio": bio, "avatar": avatar}

        params = {"user": {k: v for k, v in params.items() if v is not None}}

        if not params["user"]:
            return SpeckleException(
                message=(
                    "You must provide at least one field to update your user profile"
                )
            )

        return self.make_request(
            query=query, params=params, return_type="userUpdate", parse_response=False
        )

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def activity(
        self,
        user_id: Optional[str] = None,
        limit: int = 20,
        action_type: Optional[str] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        cursor: Optional[datetime] = None,
    ):
        """
        Get the activity from a given stream in an Activity collection.
        Step into the activity `items` for the list of activity.
        If no id argument is provided, will return the current authenticated
        user's activity (as extracted from the authorization header).

        Note: all timestamps arguments should be `datetime` of any tz as
        they will be converted to UTC ISO format strings

        user_id {str} -- the id of the user to get the activity from
        action_type {str} -- filter results to a single action type
        (eg: `commit_create` or `commit_receive`)
        limit {int} -- max number of Activity items to return
        before {datetime}
            -- latest cutoff for activity (ie: return all activity _before_ this time)
        after {datetime}
            -- oldest cutoff for activity (ie: return all activity _after_ this time)
        cursor {datetime} -- timestamp cursor for pagination
        """

        query = gql(
            """
            query UserActivity(
                $user_id: String,
                $action_type: String,
                $before:DateTime,
                $after: DateTime,
                $cursor: DateTime,
                $limit: Int
                ){
                user(id: $user_id) {
                    activity(
                        actionType: $action_type,
                        before: $before,
                        after: $after,
                        cursor: $cursor,
                        limit: $limit
                        ) {
                        totalCount
                        cursor
                        items {
                            actionType
                            info
                            userId
                            streamId
                            resourceId
                            resourceType
                            message
                            time
                        }
                    }
                }
            }
            """
        )

        params = {
            "user_id": user_id,
            "limit": limit,
            "action_type": action_type,
            "before": before.astimezone(timezone.utc).isoformat() if before else before,
            "after": after.astimezone(timezone.utc).isoformat() if after else after,
            "cursor": cursor.astimezone(timezone.utc).isoformat() if cursor else cursor,
        }

        return self.make_request(
            query=query,
            params=params,
            return_type=["user", "activity"],
            schema=ActivityCollection,
        )

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def get_all_pending_invites(self) -> List[PendingStreamCollaborator]:
        """Get all of the active user's pending stream invites

        Requires Speckle Server version >= 2.6.4

        Returns:
            List[PendingStreamCollaborator]
                -- a list of pending invites for the current user
        """
        self._check_invites_supported()

        query = gql(
            """
            query StreamInvites {
                streamInvites{
                    id
                    token
                    inviteId
                    streamId
                    streamName
                    title
                    role
                    invitedBy {
                        id
                        name
                        company
                        avatar
                    }
                }
            }
            """
        )

        return self.make_request(
            query=query,
            return_type="streamInvites",
            schema=PendingStreamCollaborator,
        )

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def get_pending_invite(
        self, stream_id: str, token: Optional[str] = None
    ) -> Optional[PendingStreamCollaborator]:
        """Get a particular pending invite for the active user on a given stream.
        If no invite_id is provided, any valid invite will be returned.

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to look for invites on
            token {str} -- the token of the invite to look for (optional)

        Returns:
            PendingStreamCollaborator
                -- the invite for the given stream (or None if it isn't found)
        """
        self._check_invites_supported()

        query = gql(
            """
            query StreamInvite($streamId: String!, $token: String) {
                streamInvite(streamId: $streamId, token: $token) {
                    id
                    token
                    streamId
                    streamName
                    title
                    role
                    invitedBy {
                        id
                        name
                        company
                        avatar
                    }
                }
            }
            """
        )

        params = {"streamId": stream_id}
        if token:
            params["token"] = token

        return self.make_request(
            query=query,
            params=params,
            return_type="streamInvite",
            schema=PendingStreamCollaborator,
        )



from typing import Optional
from enum import Enum

from specklepy.objects.base import Base
from specklepy.objects.geometry import Plane


class AxisType(int, Enum):
    Cartesian = 0
    Cylindrical = 1
    Spherical = 2


class Axis(Base, speckle_type="Objects.Structural.Geometry.Axis"):
    name: Optional[str] = None
    axisType: Optional[AxisType] = None
    plane: Optional[Plane] = None



from dataclasses import dataclass
from enum import Enum
from unicodedata import name


class HostAppVersion(Enum):
    v = "v"
    v6 = "v6"
    v7 = "v7"
    v2019 = "v2019"
    v2020 = "v2020"
    v2021 = "v2021"
    v2022 = "v2022"
    v2023 = "v2023"
    v2024 = "v2024"
    v2025 = "v2025"
    vSandbox = "vSandbox"
    vRevit = "vRevit"
    vRevit2021 = "vRevit2021"
    vRevit2022 = "vRevit2022"
    vRevit2023 = "vRevit2023"
    vRevit2024 = "vRevit2024"
    vRevit2025 = "vRevit2025"
    v25 = "v25"
    v26 = "v26"

    def __repr__(self) -> str:
        return self.value

    def __str__(self) -> str:
        return self.value


@dataclass
class HostApplication:
    name: str
    slug: str

    def get_version(self, version: HostAppVersion) -> str:
        return f"{name.replace(' ', '')}{str(version).strip('v')}"


RHINO = HostApplication("Rhino", "rhino")
GRASSHOPPER = HostApplication("Grasshopper", "grasshopper")
REVIT = HostApplication("Revit", "revit")
DYNAMO = HostApplication("Dynamo", "dynamo")
UNITY = HostApplication("Unity", "unity")
GSA = HostApplication("GSA", "gsa")
CIVIL = HostApplication("Civil 3D", "civil3d")
AUTOCAD = HostApplication("AutoCAD", "autocad")
MICROSTATION = HostApplication("MicroStation", "microstation")
OPENROADS = HostApplication("OpenRoads", "openroads")
OPENRAIL = HostApplication("OpenRail", "openrail")
OPENBUILDINGS = HostApplication("OpenBuildings", "openbuildings")
ETABS = HostApplication("ETABS", "etabs")
SAP2000 = HostApplication("SAP2000", "sap2000")
CSIBRIDGE = HostApplication("CSIBridge", "csibridge")
SAFE = HostApplication("SAFE", "safe")
TEKLASTRUCTURES = HostApplication("Tekla Structures", "teklastructures")
DXF = HostApplication("DXF Converter", "dxf")
EXCEL = HostApplication("Excel", "excel")
UNREAL = HostApplication("Unreal", "unreal")
POWERBI = HostApplication("Power BI", "powerbi")
BLENDER = HostApplication("Blender", "blender")
QGIS = HostApplication("QGIS", "qgis")
ARCGIS = HostApplication("ArcGIS", "arcgis")
SKETCHUP = HostApplication("SketchUp", "sketchup")
ARCHICAD = HostApplication("Archicad", "archicad")
TOPSOLID = HostApplication("TopSolid", "topsolid")
PYTHON = HostApplication("Python", "python")
NET = HostApplication(".NET", "net")
OTHER = HostApplication("Other", "other")

_app_name_host_app_mapping = {
    "dynamo": DYNAMO,
    "revit": REVIT,
    "autocad": AUTOCAD,
    "civil": CIVIL,
    "rhino": RHINO,
    "grasshopper": GRASSHOPPER,
    "unity": UNITY,
    "gsa": GSA,
    "microstation": MICROSTATION,
    "openroads": OPENROADS,
    "openrail": OPENRAIL,
    "openbuildings": OPENBUILDINGS,
    "etabs": ETABS,
    "sap": SAP2000,
    "csibridge": CSIBRIDGE,
    "safe": SAFE,
    "teklastructures": TEKLASTRUCTURES,
    "dxf": DXF,
    "excel": EXCEL,
    "unreal": UNREAL,
    "powerbi": POWERBI,
    "blender": BLENDER,
    "qgis": QGIS,
    "arcgis": ARCGIS,
    "sketchup": SKETCHUP,
    "archicad": ARCHICAD,
    "topsolid": TOPSOLID,
    "python": PYTHON,
    "net": NET,
}


def get_host_app_from_string(app_name: str) -> HostApplication:
    app_name = app_name.lower().replace(" ", "")
    for partial_app_name, host_app in _app_name_host_app_mapping.items():
        if partial_app_name in app_name:
            return host_app
    return HostApplication(app_name, app_name)


if __name__ == "__main__":
    print(HostAppVersion.v)



from enum import Enum
from typing import Any, List, Optional

from specklepy.objects.base import Base
from specklepy.objects.encoding import CurveArray, CurveTypeEncoding, ObjectArray
from specklepy.objects.primitive import Interval
from specklepy.objects.units import get_encoding_from_units, get_units_from_encoding

GEOMETRY = "Objects.Geometry."


class Point(Base, speckle_type=GEOMETRY + "Point"):
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(x: {self.x}, y: {self.y}, z: {self.z}, id:"
            f" {self.id}, speckle_type: {self.speckle_type})"
        )

    @classmethod
    def from_list(cls, args: List[float]) -> "Point":
        """
        Create a new Point from a list of three floats
        representing the x, y, and z coordinates
        """
        return cls(x=args[0], y=args[1], z=args[2])

    def to_list(self) -> List[Any]:
        return [self.x, self.y, self.z]

    @classmethod
    def from_coords(cls, x: float = 0.0, y: float = 0.0, z: float = 0.0):
        """Create a new Point from x, y, and z values"""
        pt = Point()
        pt.x, pt.y, pt.z = x, y, z
        return pt


class Pointcloud(
    Base,
    speckle_type=GEOMETRY + "Pointcloud",
    chunkable={"points": 31250, "colors": 62500, "sizes": 62500},
):
    points: Optional[List[float]] = None
    colors: Optional[List[int]] = None
    sizes: Optional[List[float]] = None
    bbox: Optional["Box"] = None


class Vector(Base, speckle_type=GEOMETRY + "Vector"):
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    applicationId: Optional[str] = None

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__} "
            "(x: {self.x}, y: {self.y}, z: {self.z}, id: {self.id}, "
            "speckle_type: {self.speckle_type})"
        )

    @classmethod
    def from_list(cls, args: List[float]) -> "Vector":
        """
        Create from a list of three floats representing the x, y, and z coordinates.
        """
        return cls(x=args[0], y=args[1], z=args[2])

    def to_list(self) -> List[float]:
        return [self.x, self.y, self.z]

    @classmethod
    def from_coords(cls, x: float = 0.0, y: float = 0.0, z: float = 0.0) -> "Vector":
        """Create a new Point from x, y, and z values"""
        v = Vector()
        v.x, v.y, v.z = x, y, z
        return v


class ControlPoint(Point, speckle_type=GEOMETRY + "ControlPoint"):
    weight: Optional[float] = None


class Plane(Base, speckle_type=GEOMETRY + "Plane"):
    origin: Point = Point()
    normal: Vector = Vector()
    xdir: Vector = Vector()
    ydir: Vector = Vector()

    @classmethod
    def from_list(cls, args: List[Any]) -> "Plane":
        return cls(
            origin=Point.from_list(args[:3]),
            normal=Vector.from_list(args[3:6]),
            xdir=Vector.from_list(args[6:9]),
            ydir=Vector.from_list(args[9:12]),
            units=get_units_from_encoding(args[-1]),
        )

    def to_list(self) -> List[Any]:
        return [
            *self.origin.to_list(),
            *self.normal.to_list(),
            *self.xdir.to_list(),
            *self.ydir.to_list(),
            get_encoding_from_units(self._units),
        ]


class Box(Base, speckle_type=GEOMETRY + "Box"):
    basePlane: Plane = Plane()
    xSize: Interval = Interval()
    ySize: Interval = Interval()
    zSize: Interval = Interval()
    area: Optional[float] = None
    volume: Optional[float] = None


class Line(Base, speckle_type=GEOMETRY + "Line"):
    start: Point = Point()
    end: Optional[Point] = None
    domain: Optional[Interval] = None
    bbox: Optional[Box] = None
    length: Optional[float] = None

    @classmethod
    def from_list(cls, args: List[Any]) -> "Line":
        return cls(
            start=Point.from_list(args[1:4]),
            end=Point.from_list(args[4:7]),
            domain=Interval.from_list(args[7:10]),
            units=get_units_from_encoding(args[-1]),
        )

    def to_list(self) -> List[Any]:
        domain = self.domain.to_list() if self.domain else [0, 1]
        return [
            CurveTypeEncoding.Line.value,
            *self.start.to_list(),
            *self.end.to_list(),
            *domain,
            get_encoding_from_units(self._units),
        ]


class Arc(Base, speckle_type=GEOMETRY + "Arc"):
    radius: Optional[float] = None
    startAngle: Optional[float] = None
    endAngle: Optional[float] = None
    angleRadians: Optional[float] = None
    plane: Optional[Plane] = None
    domain: Optional[Interval] = None
    startPoint: Optional[Point] = None
    midPoint: Optional[Point] = None
    endPoint: Optional[Point] = None
    bbox: Optional[Box] = None
    area: Optional[float] = None
    length: Optional[float] = None

    @classmethod
    def from_list(cls, args: List[Any]) -> "Arc":
        return cls(
            radius=args[1],
            startAngle=args[2],
            endAngle=args[3],
            angleRadians=args[4],
            domain=Interval.from_list(args[5:7]),
            plane=Plane.from_list(args[7:20]),
            startPoint=Point.from_list(args[20:23]),
            midPoint=Point.from_list(args[23:26]),
            endPoint=Point.from_list(args[26:29]),
            units=get_units_from_encoding(args[-1]),
        )

    def to_list(self) -> List[Any]:
        return [
            CurveTypeEncoding.Arc.value,
            self.radius,
            self.startAngle,
            self.endAngle,
            self.angleRadians,
            *self.domain.to_list(),
            *self.plane.to_list(),
            *self.startPoint.to_list(),
            *self.midPoint.to_list(),
            *self.endPoint.to_list(),
            get_encoding_from_units(self._units),
        ]


class Circle(Base, speckle_type=GEOMETRY + "Circle"):
    radius: Optional[float] = None
    plane: Optional[Plane] = None
    domain: Optional[Interval] = None
    bbox: Optional[Box] = None
    area: Optional[float] = None
    length: Optional[float] = None

    @classmethod
    def from_list(cls, args: List[Any]) -> "Circle":
        return cls(
            radius=args[1],
            domain=Interval.from_list(args[2:4]),
            plane=Plane.from_list(args[4:17]),
            units=get_units_from_encoding(args[-1]),
        )

    def to_list(self) -> List[Any]:
        return [
            CurveTypeEncoding.Circle.value,
            self.radius,
            *self.domain.to_list(),
            *self.plane.to_list(),
            get_encoding_from_units(self._units),
        ]


class Ellipse(Base, speckle_type=GEOMETRY + "Ellipse"):
    firstRadius: Optional[float] = None
    secondRadius: Optional[float] = None
    plane: Optional[Plane] = None
    domain: Optional[Interval] = None
    trimDomain: Optional[Interval] = None
    bbox: Optional[Box] = None
    area: Optional[float] = None
    length: Optional[float] = None

    @classmethod
    def from_list(cls, args: List[Any]) -> "Ellipse":
        return cls(
            firstRadius=args[1],
            secondRadius=args[2],
            domain=Interval.from_list(args[3:5]),
            plane=Plane.from_list(args[5:18]),
            units=get_units_from_encoding(args[-1]),
        )

    def to_list(self) -> List[Any]:
        return [
            CurveTypeEncoding.Ellipse.value,
            self.firstRadius,
            self.secondRadius,
            *self.domain.to_list(),
            *self.plane.to_list(),
            get_encoding_from_units(self._units),
        ]


class Polyline(Base, speckle_type=GEOMETRY + "Polyline", chunkable={"value": 20000}):
    value: Optional[List[float]] = None
    closed: Optional[bool] = None
    domain: Optional[Interval] = None
    bbox: Optional[Box] = None
    area: Optional[float] = None
    length: Optional[float] = None

    @classmethod
    def from_points(cls, points: List[Point]):
        """Create a new Polyline from a list of Points"""
        polyline = cls()
        polyline.units = points[0].units
        polyline.value = []
        for point in points:
            polyline.value.extend([point.x, point.y, point.z])
        return polyline

    @classmethod
    def from_list(cls, args: List[Any]) -> "Polyline":
        point_count = args[4]
        return cls(
            closed=bool(args[1]),
            domain=Interval.from_list(args[2:4]),
            value=args[5 : 5 + point_count],
            units=get_units_from_encoding(args[-1]),
        )

    def to_list(self) -> List[Any]:
        return [
            CurveTypeEncoding.Polyline.value,
            int(self.closed),
            *self.domain.to_list(),
            len(self.value),
            *self.value,
            get_encoding_from_units(self._units),
        ]

    def as_points(self) -> List[Point]:
        """Converts the `value` attribute to a list of Points"""
        if not self.value:
            return

        if len(self.value) % 3:
            raise ValueError("Points array malformed: length%3 != 0.")

        values = iter(self.value)
        return [
            Point(x=v, y=next(values), z=next(values), units=self.units) for v in values
        ]


class SpiralType(Enum):
    Biquadratic = (0,)
    BiquadraticParabola = (1,)
    Bloss = (2,)
    Clothoid = (3,)
    Cosine = (4,)
    Cubic = (5,)
    CubicParabola = (6,)
    Radioid = (7,)
    Sinusoid = (8,)
    Unknown = 9


class Spiral(Base, speckle_type=GEOMETRY + "Spiral", detachable={"displayValue"}):
    startPoint: Optional[Point] = None
    endPoint: Optional[Point]
    plane: Optional[Plane]
    turns: Optional[float]
    pitchAxis: Optional[Vector] = Vector()
    pitch: float = 0
    spiralType: Optional[SpiralType] = None
    displayValue: Optional[Polyline] = None
    bbox: Optional[Box] = None
    length: Optional[float] = None
    domain: Optional[Interval] = None


class Curve(
    Base,
    speckle_type=GEOMETRY + "Curve",
    chunkable={"points": 20000, "weights": 20000, "knots": 20000},
):
    degree: Optional[int] = None
    periodic: Optional[bool] = None
    rational: Optional[bool] = None
    points: Optional[List[float]] = None
    weights: Optional[List[float]] = None
    knots: Optional[List[float]] = None
    domain: Optional[Interval] = None
    displayValue: Optional[Polyline] = None
    closed: Optional[bool] = None
    bbox: Optional[Box] = None
    area: Optional[float] = None
    length: Optional[float] = None

    def as_points(self) -> List[Point]:
        """Converts the `value` attribute to a list of Points"""
        if not self.points:
            return

        if len(self.points) % 3:
            raise ValueError("Points array malformed: length%3 != 0.")

        values = iter(self.points)
        return [
            Point(x=v, y=next(values), z=next(values), units=self.units) for v in values
        ]

    @classmethod
    def from_list(cls, args: List[Any]) -> "Curve":
        point_count = int(args[7])
        weights_count = int(args[8])
        knots_count = int(args[9])

        points_start = 10
        weights_start = 10 + point_count
        knots_start = weights_start + weights_count
        knots_end = knots_start + knots_count

        return cls(
            degree=int(args[1]),
            periodic=bool(args[2]),
            rational=bool(args[3]),
            closed=bool(args[4]),
            domain=Interval.from_list(args[5:7]),
            points=args[points_start:weights_start],
            weights=args[weights_start:knots_start],
            knots=args[knots_start:knots_end],
            units=get_units_from_encoding(args[-1]),
        )

    def to_list(self) -> List[Any]:
        return [
            CurveTypeEncoding.Curve.value,
            self.degree,
            int(self.periodic),
            int(self.rational),
            int(self.closed),
            *self.domain.to_list(),
            len(self.points),
            len(self.weights),
            len(self.knots),
            *self.points,
            *self.weights,
            *self.knots,
            get_encoding_from_units(self._units),
        ]


class Polycurve(Base, speckle_type=GEOMETRY + "Polycurve"):
    segments: Optional[List[Base]] = None
    domain: Optional[Interval] = None
    closed: Optional[bool] = None
    bbox: Optional[Box] = None
    area: Optional[float] = None
    length: Optional[float] = None

    @classmethod
    def from_list(cls, args: List[Any]) -> "Polycurve":
        curve_arrays = CurveArray(args[5:-1])
        return cls(
            closed=bool(args[1]),
            domain=Interval.from_list(args[2:4]),
            segments=curve_arrays.to_curves(),
            units=get_units_from_encoding(args[-1]),
        )

    def to_list(self) -> List[Any]:
        curve_array = CurveArray.from_curves(self.segments).data
        return [
            CurveTypeEncoding.Polycurve.value,
            int(self.closed),
            *self.domain.to_list(),
            len(curve_array),
            *curve_array,
            get_encoding_from_units(self._units),
        ]


class Extrusion(Base, speckle_type=GEOMETRY + "Extrusion"):
    capped: Optional[bool] = None
    profile: Optional[Base] = None
    pathStart: Optional[Point] = None
    pathEnd: Optional[Point] = None
    pathCurve: Optional[Base] = None
    pathTangent: Optional[Base] = None
    profiles: Optional[List[Base]] = None
    length: Optional[float] = None
    area: Optional[float] = None
    volume: Optional[float] = None
    bbox: Optional[Box] = None


class Mesh(
    Base,
    speckle_type=GEOMETRY + "Mesh",
    chunkable={
        "vertices": 2000,
        "faces": 2000,
        "colors": 2000,
        "textureCoordinates": 2000,
    },
):
    vertices: Optional[List[float]] = None
    faces: Optional[List[int]] = None
    colors: Optional[List[int]] = None
    textureCoordinates: Optional[List[float]] = None
    bbox: Optional[Box] = None
    area: Optional[float] = None
    volume: Optional[float] = None

    @classmethod
    def create(
        cls,
        vertices: List[float],
        faces: List[int],
        colors: Optional[List[int]] = None,
        texture_coordinates: Optional[List[float]] = None,
    ) -> "Mesh":
        """
        Create a new Mesh from lists representing its vertices, faces,
        colors (optional), and texture coordinates (optional).

        This will initialise empty lists for colors and texture coordinates
        if you do not provide any.
        """
        return cls(
            vertices=vertices,
            faces=faces,
            colors=colors or [],
            textureCoordinates=texture_coordinates or [],
        )


class Surface(Base, speckle_type=GEOMETRY + "Surface"):
    degreeU: Optional[int] = None
    degreeV: Optional[int] = None
    rational: Optional[bool] = None
    area: Optional[float] = None
    pointData: Optional[List[float]] = None
    countU: Optional[int] = None
    countV: Optional[int] = None
    bbox: Optional[Box] = None
    closedU: Optional[bool] = None
    closedV: Optional[bool] = None
    domainU: Optional[Interval] = None
    domainV: Optional[Interval] = None
    knotsU: Optional[List[float]] = None
    knotsV: Optional[List[float]] = None

    @classmethod
    def from_list(cls, args: List[Any]) -> "Surface":
        point_count = int(args[11])
        knots_u_count = int(args[12])
        knots_v_count = int(args[13])

        start_point_data = 14
        start_knots_u = start_point_data + point_count
        start_knots_v = start_knots_u + knots_u_count

        return cls(
            degreeU=int(args[0]),
            degreeV=int(args[1]),
            countU=int(args[2]),
            countV=int(args[3]),
            rational=bool(args[4]),
            closedU=bool(args[5]),
            closedV=bool(args[6]),
            domainU=Interval(start=args[7], end=args[8]),
            domainV=Interval(start=args[9], end=args[10]),
            pointData=args[start_point_data:start_knots_u],
            knotsU=args[start_knots_u:start_knots_v],
            knotsV=args[start_knots_v : start_knots_v + knots_v_count],
            units=get_units_from_encoding(args[-1]),
        )

    def to_list(self) -> List[Any]:
        return [
            self.degreeU,
            self.degreeV,
            self.countU,
            self.countV,
            int(self.rational),
            int(self.closedU),
            int(self.closedV),
            *self.domainU.to_list(),
            *self.domainV.to_list(),
            len(self.pointData),
            len(self.knotsU),
            len(self.knotsV),
            *self.pointData,
            *self.knotsU,
            *self.knotsV,
            get_encoding_from_units(self._units),
        ]


class BrepFace(Base, speckle_type=GEOMETRY + "BrepFace"):
    _Brep: Optional["Brep"] = None
    SurfaceIndex: Optional[int] = None
    OuterLoopIndex: Optional[int] = None
    OrientationReversed: Optional[bool] = None
    LoopIndices: Optional[List[int]] = None

    @property
    def _outer_loop(self):
        return self._Brep.Loops[self.OuterLoopIndex]  # pylint: disable=no-member

    @property
    def _surface(self):
        return self._Brep.Surfaces[self.SurfaceIndex]  # pylint: disable=no-member

    @property
    def _loops(self):
        if self.LoopIndices:
            # pylint: disable=not-an-iterable, no-member
            return [self._Brep.Loops[i] for i in self.LoopIndices]

    @classmethod
    def from_list(cls, args: List[Any], brep: "Brep" = None) -> "BrepFace":
        return cls(
            _Brep=brep,
            SurfaceIndex=args[0],
            OuterLoopIndex=args[1],
            OrientationReversed=bool(args[2]),
            LoopIndices=args[3:],
        )

    def to_list(self) -> List[Any]:
        return [
            self.SurfaceIndex,
            self.OuterLoopIndex,
            int(self.OrientationReversed),
            *self.LoopIndices,
        ]


class BrepEdge(Base, speckle_type=GEOMETRY + "BrepEdge"):
    _Brep: Optional["Brep"] = None
    Curve3dIndex: Optional[int] = None
    TrimIndices: Optional[List[int]] = None
    StartIndex: Optional[int] = None
    EndIndex: Optional[int] = None
    ProxyCurveIsReversed: Optional[bool] = None
    Domain: Optional[Interval] = None

    @property
    def _start_vertex(self):
        return self._Brep.Vertices[self.StartIndex]

    @property
    def _end_vertex(self):
        return self._Brep.Vertices[self.EndIndex]

    @property
    def _trims(self):
        if self.TrimIndices:
            # pylint: disable=not-an-iterable
            return [self._Brep.Trims[i] for i in self.TrimIndices]

    @property
    def _curve(self):
        return self._Brep.Curve3D[self.Curve3dIndex]

    @classmethod
    def from_list(cls, args: List[Any], brep: "Brep" = None) -> "BrepEdge":
        domain_start = args[4]
        domain_end = args[5]
        domain = (
            Interval(start=domain_start, end=domain_end)
            if None not in (domain_start, domain_end)
            else None
        )
        return cls(
            _Brep=brep,
            Curve3dIndex=int(args[0]),
            TrimIndices=[int(t) for t in args[6:]],
            StartIndex=int(args[1]),
            EndIndex=int(args[2]),
            ProxyCurveIsReversed=bool(args[3]),
            Domain=domain,
        )

    def to_list(self) -> List[Any]:
        return [
            self.Curve3dIndex,
            self.StartIndex,
            self.EndIndex,
            int(self.ProxyCurveIsReversed),
            self.Domain.start,
            self.Domain.end,
            *self.TrimIndices,
        ]


class BrepLoopType(int, Enum):
    Unknown = 0
    Outer = 1
    Inner = 2
    Slit = 3
    CurveOnSurface = 4
    PointOnSurface = 5


class BrepLoop(Base, speckle_type=GEOMETRY + "BrepLoop"):
    _Brep: Optional["Brep"] = None
    FaceIndex: Optional[Optional[int]] = None
    TrimIndices: Optional[List[int]] = None
    Type: Optional[BrepLoopType] = None

    @property
    def _face(self):
        return self._Brep.Faces[self.FaceIndex]

    @property
    def _trims(self):
        if self.TrimIndices:
            # pylint: disable=not-an-iterable
            return [self._Brep.Trims[i] for i in self.TrimIndices]

    @classmethod
    def from_list(cls, args: List[any], brep: "Brep" = None):
        return cls(
            _Brep=brep,
            FaceIndex=args[0],
            Type=BrepLoopType(args[1]),
            TrimIndices=args[2:],
        )

    def to_list(self) -> List[int]:
        return [
            self.FaceIndex,
            self.Type.value,
            *self.TrimIndices,
        ]


class BrepTrimType(int, Enum):
    Unknown = 0
    Boundary = 1
    Mated = 2
    Seam = 3
    Singular = 4
    CurveOnSurface = 5
    PointOnSurface = 6
    Slit = 7


class BrepTrim(Base, speckle_type=GEOMETRY + "BrepTrim"):
    _Brep: Optional["Brep"] = None
    EdgeIndex: Optional[int] = None
    StartIndex: Optional[int] = None
    EndIndex: Optional[int] = None
    FaceIndex: Optional[int] = None
    LoopIndex: Optional[int] = None
    CurveIndex: Optional[int] = None
    IsoStatus: Optional[int] = None
    TrimType: Optional[BrepTrimType] = None
    IsReversed: Optional[bool] = None
    Domain: Optional[Interval] = None

    @property
    def _face(self):
        if self._Brep:
            return self._Brep.Faces[self.FaceIndex]  # pylint: disable=no-member

    @property
    def _loop(self):
        if self._Brep:
            return self._Brep.Loops[self.LoopIndex]  # pylint: disable=no-member

    @property
    def _edge(self):
        if self._Brep:
            # pylint: disable=no-member
            return self._Brep.Edges[self.EdgeIndex] if self.EdgeIndex != -1 else None

    @property
    def _curve_2d(self):
        if self._Brep:
            return self._Brep.Curve2D[self.CurveIndex]  # pylint: disable=no-member

    @classmethod
    def from_list(cls, args: List[Any], brep: "Brep" = None) -> "BrepTrim":
        return cls(
            _Brep=brep,
            EdgeIndex=args[0],
            StartIndex=args[1],
            EndIndex=args[2],
            FaceIndex=args[3],
            LoopIndex=args[4],
            CurveIndex=args[5],
            IsoStatus=args[6],
            TrimType=BrepTrimType(args[7]),
            IsReversed=bool(args[8]),
        )

    def to_list(self) -> List[Any]:
        return [
            self.EdgeIndex,
            self.StartIndex,
            self.EndIndex,
            self.FaceIndex,
            self.LoopIndex,
            self.CurveIndex,
            self.IsoStatus,
            self.TrimType.value,
            int(self.IsReversed),
        ]


class Brep(
    Base,
    speckle_type=GEOMETRY + "Brep",
    chunkable={
        "SurfacesValue": 31250,
        "Curve3DValues": 31250,
        "Curve2DValues": 31250,
        "VerticesValue": 31250,
        "EdgesValue": 62500,
        "LoopsValue": 62500,
        "FacesValue": 62500,
        "TrimsValue": 62500,
    },
    detachable={"displayValue"},
    serialize_ignore={
        "Surfaces",
        "Curve3D",
        "Curve2D",
        "Vertices",
        "Trims",
        "Edges",
        "Loops",
        "Faces",
    },
):
    provenance: Optional[str] = None
    bbox: Optional[Box] = None
    area: Optional[float] = None
    volume: Optional[float] = None
    _displayValue: Optional[List[Mesh]] = None
    Surfaces: Optional[List[Surface]] = None
    Curve3D: Optional[List[Base]] = None
    Curve2D: Optional[List[Base]] = None
    Vertices: Optional[List[Point]] = None
    Edges: Optional[List[BrepEdge]] = None
    Loops: Optional[List[BrepLoop]] = None
    Faces: Optional[List[BrepFace]] = None
    Trims: Optional[List[BrepTrim]] = None
    IsClosed: Optional[bool] = None
    Orientation: Optional[int] = None

    def _inject_self_into_children(self, children: Optional[List[Base]]) -> List[Base]:
        if children is None:
            return children

        for child in children:
            child._Brep = self  # pylint: disable=protected-access
        return children

    # set as prop for now for backwards compatibility
    @property
    def displayValue(self) -> List[Mesh]:
        return self._displayValue

    @displayValue.setter
    def displayValue(self, value):
        if isinstance(value, Mesh):
            self._displayValue = [value]
        elif isinstance(value, list):
            self._displayValue = value

    @property
    def EdgesValue(self) -> List[BrepEdge]:
        return None if self.Edges is None else ObjectArray.from_objects(self.Edges).data

    @EdgesValue.setter
    def EdgesValue(self, value: List[float]):
        if not value:
            return

        self.Edges = ObjectArray.decode_data(value, BrepEdge.from_list, brep=self)

    @property
    def LoopsValue(self) -> List[BrepLoop]:
        return None if self.Loops is None else ObjectArray.from_objects(self.Loops).data

    @LoopsValue.setter
    def LoopsValue(self, value: List[int]):
        if not value:
            return

        self.Loops = ObjectArray.decode_data(value, BrepLoop.from_list, brep=self)

    @property
    def FacesValue(self) -> List[int]:
        return None if self.Faces is None else ObjectArray.from_objects(self.Faces).data

    @FacesValue.setter
    def FacesValue(self, value: List[int]):
        if not value:
            return

        self.Faces = ObjectArray.decode_data(value, BrepFace.from_list, brep=self)

    @property
    def SurfacesValue(self) -> List[float]:
        return (
            None
            if self.Surfaces is None
            else ObjectArray.from_objects(self.Surfaces).data
        )

    @SurfacesValue.setter
    def SurfacesValue(self, value: List[float]):
        if not value:
            return

        self.Surfaces = ObjectArray.decode_data(value, Surface.from_list)

    @property
    def Curve3DValues(self) -> List[float]:
        return (
            None if self.Curve3D is None else CurveArray.from_curves(self.Curve3D).data
        )

    @Curve3DValues.setter
    def Curve3DValues(self, value: List[float]):
        crv_array = CurveArray(value)
        self.Curve3D = crv_array.to_curves()

    @property
    def Curve2DValues(self) -> List[Base]:
        return (
            None if self.Curve2D is None else CurveArray.from_curves(self.Curve2D).data
        )

    @Curve2DValues.setter
    def Curve2DValues(self, value: List[float]):
        crv_array = CurveArray(value)
        self.Curve2D = crv_array.to_curves()

    @property
    def VerticesValue(self) -> List[Point]:
        if self.Vertices is None:
            return None
        encoded_unit = get_encoding_from_units(self.Vertices[0].units)
        values = [encoded_unit]
        for vertex in self.Vertices:
            values.extend(vertex.to_list())
        return values

    @VerticesValue.setter
    def VerticesValue(self, value: List[float]):
        value = value.copy()
        units = get_units_from_encoding(value.pop(0))

        vertices = []

        for i in range(0, len(value), 3):
            vertex = Point.from_list(value[i : i + 3])
            vertex.units = units
            vertices.append(vertex)

        self.Vertices = vertices

    # TODO: can this be consistent with loops, edges, faces, curves, etc and prepend with the chunk list? needs to happen in sharp first
    @property
    def TrimsValue(self) -> List[float]:
        # return None if self.Trims is None else ObjectArray.from_objects(self.Trims).data
        if not self.Trims:
            return
        value = []
        for trim in self.Trims:
            value.extend(trim.to_list())
        return value

    @TrimsValue.setter
    def TrimsValue(self, value: List[float]):
        if not value:
            return

        # self.Trims = ObjectArray.decode_data(value, BrepTrim.from_list, brep=self)
        self.Trims = [
            BrepTrim.from_list(value[i : i + 9], self) for i in range(0, len(value), 9)
        ]


BrepEdge.update_forward_refs()
BrepLoop.update_forward_refs()
BrepTrim.update_forward_refs()
BrepFace.update_forward_refs()



"""
Provides uniform and consistent path helpers for `specklepy`
"""
import os
import sys
from pathlib import Path
from typing import Optional

from specklepy.logging.exceptions import SpeckleException

_user_data_env_var = "SPECKLE_USERDATA_PATH"


def _path() -> Optional[Path]:
    """Read the user data path override setting."""
    path_override = os.environ.get(_user_data_env_var)
    if path_override:
        return Path(path_override)
    return None


_application_name = "Speckle"


def override_application_name(application_name: str) -> None:
    """Override the global Speckle application name."""
    global _application_name
    _application_name = application_name


def override_application_data_path(path: Optional[str]) -> None:
    """
    Override the global Speckle application data path.

    If the value of path is `None` the environment variable gets deleted.
    """
    if path:
        os.environ[_user_data_env_var] = path
    else:
        os.environ.pop(_user_data_env_var, None)


_blob_folder_name = "Blobs"


def override_blob_storage_folder(blob_folder_name: str) -> None:
    """Override the global Blob storage folder name."""
    global _blob_folder_name
    _blob_folder_name = blob_folder_name


_accounts_folder_name = "Accounts"


def override_accounts_folder_name(accounts_folder_name: str) -> None:
    """Override the global Accounts folder name."""
    global _accounts_folder_name
    _accounts_folder_name = accounts_folder_name


_objects_folder_name = "Objects"


def override_objects_folder_name(objects_folder_name: str) -> None:
    """Override global Objects folder name."""
    global _objects_folder_name
    _objects_folder_name = objects_folder_name


def _ensure_folder_exists(base_path: Path, folder_name: str) -> Path:
    path = base_path.joinpath(folder_name)
    path.mkdir(exist_ok=True, parents=True)
    return path


def user_application_data_path() -> Path:
    """Get the platform specific user configuration folder path"""
    path_override = _path()
    if path_override:
        return path_override

    try:
        if sys.platform.startswith("win"):
            app_data_path = os.getenv("APPDATA")
            if not app_data_path:
                raise SpeckleException(
                    message="Cannot get appdata path from environment."
                )
            return Path(app_data_path)
        else:
            # try getting the standard XDG_DATA_HOME value
            # as that is used as an override
            app_data_path = os.getenv("XDG_DATA_HOME")
            if app_data_path:
                return Path(app_data_path)
            else:
                return _ensure_folder_exists(Path.home(), ".config")
    except Exception as ex:
        raise SpeckleException(
            message="Failed to initialize user application data path.", exception=ex
        )


def user_speckle_folder_path() -> Path:
    """Get the folder where the user's Speckle data should be stored."""
    return _ensure_folder_exists(user_application_data_path(), _application_name)


def user_speckle_connector_installation_path(host_application: str) -> Path:
    """
    Gets a connector specific installation folder.

    In this folder we can put our connector installation and all python packages.
    """
    return _ensure_folder_exists(
        _ensure_folder_exists(user_speckle_folder_path(), "connector_installations"),
        host_application,
    )


def accounts_folder_path() -> Path:
    """Get the folder where the Speckle accounts data should be stored."""
    return _ensure_folder_exists(user_speckle_folder_path(), _accounts_folder_name)


def blob_storage_path(path: Optional[Path] = None) -> Path:
    return _ensure_folder_exists(path or user_speckle_folder_path(), _blob_folder_name)



import contextlib
from enum import Enum
from inspect import isclass
from typing import (
    Any,
    ClassVar,
    Dict,
    ForwardRef,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    get_type_hints,
)
from warnings import warn

from stringcase import pascalcase

from specklepy.logging.exceptions import SpeckleException, SpeckleInvalidUnitException
from specklepy.objects.units import Units, get_units_from_string
from specklepy.transports.memory import MemoryTransport

PRIMITIVES = (int, float, str, bool)

# to remove from dir() when calling get_member_names()
REMOVE_FROM_DIR = {
    "Config",
    "_Base__dict_helper",
    "__annotations__",
    "__class__",
    "__delattr__",
    "__dict__",
    "__dir__",
    "__doc__",
    "__eq__",
    "__format__",
    "__ge__",
    "__getattribute__",
    "__getitem__",
    "__gt__",
    "__hash__",
    "__init__",
    "__init_subclass__",
    "__le__",
    "__lt__",
    "__module__",
    "__ne__",
    "__new__",
    "__reduce__",
    "__reduce_ex__",
    "__repr__",
    "__setattr__",
    "__setitem__",
    "__sizeof__",
    "__str__",
    "__subclasshook__",
    "__weakref__",
    "_chunk_size_default",
    "_chunkable",
    "_count_descendants",
    "_attr_types",
    "_detachable",
    "_handle_object_count",
    "_type_check",
    "_type_registry",
    "_units",
    "add_chunkable_attrs",
    "add_detachable_attrs",
    "get_children_count",
    "get_dynamic_member_names",
    "get_id",
    "get_member_names",
    "get_registered_type",
    "get_typed_member_names",
    "to_dict",
    "update_forward_refs",
    "validate_prop_name",
    "from_list",
    "to_list",
}


class _RegisteringBase:
    """
    Private Base model for Speckle types.

    This is an implementation detail, please do not use this outside this module.

    This class provides automatic registration of `speckle_type` into a global,
    (class level) registry for each subclassing type.
    The type registry is a base for accurate type based (de)serialization.
    """

    speckle_type: ClassVar[str]
    _speckle_type_override: ClassVar[Optional[str]] = None
    _speckle_namespace: ClassVar[Optional[str]] = None
    _type_registry: ClassVar[Dict[str, Type["Base"]]] = {}
    _attr_types: ClassVar[Dict[str, Type]] = {}
    # dict of chunkable props and their max chunk size
    _chunkable: Dict[str, int] = {}
    _chunk_size_default: int = 1000
    _detachable: Set[str] = set()  # list of defined detachable props
    _serialize_ignore: Set[str] = set()

    @classmethod
    def get_registered_type(cls, speckle_type: str) -> Optional[Type["Base"]]:
        """Get the registered type from the protected mapping via the `speckle_type`"""
        for full_name in reversed(speckle_type.split(":")):
            maybe_type = cls._type_registry.get(full_name, None)
            if maybe_type:
                return maybe_type
        return None

    @classmethod
    def _determine_speckle_type(cls) -> str:
        """
        This method brings the speckle_type construction in par with peckle-sharp/Core.

        The implementation differs, because in Core the basis of the speckle_type if
        type.FullName, which includes the dotnet namespace name too.
        Copying that behavior is hard in python, where the concept of namespaces
        means something entirely different.

        So we enabled a speckle_type override mechanism, that enables
        """
        base_name = "Base"
        if cls.__name__ == base_name:
            return base_name

        bases = [
            b._full_name()
            for b in reversed(cls.mro())
            if issubclass(b, Base) and b.__name__ != base_name
        ]
        return ":".join(bases)

    @classmethod
    def _full_name(cls) -> str:
        base_name = "Base"
        if cls.__name__ == base_name:
            return base_name

        if cls._speckle_type_override:
            return cls._speckle_type_override

        # convert the module names to PascalCase to match c# namespace naming convention
        # also drop specklepy from the beginning
        namespace = ".".join(
            pascalcase(m)
            for m in filter(lambda name: name != "specklepy", cls.__module__.split("."))
        )
        return f"{namespace}.{cls.__name__}"

    def __init_subclass__(
        cls,
        speckle_type: Optional[str] = None,
        chunkable: Optional[Dict[str, int]] = None,
        detachable: Optional[Set[str]] = None,
        serialize_ignore: Optional[Set[str]] = None,
        **kwargs: Dict[str, Any],
    ):
        """
        Hook into subclass type creation.

        This is provides a mechanism to hook into the event of the subclass type object
        initialization. This is reused to register each subclassing type into a class
        level dictionary.
        """
        cls._speckle_type_override = speckle_type
        cls.speckle_type = cls._determine_speckle_type()
        if cls._full_name() in cls._type_registry:
            raise ValueError(
                f"The speckle_type: {speckle_type} is already registered for type: "
                f"{cls._type_registry[cls._full_name()].__name__}. "
                "Please choose a different type name."
            )
        cls._type_registry[cls._full_name()] = cls  # type: ignore
        try:
            cls._attr_types = get_type_hints(cls)
        except Exception:
            cls._attr_types = getattr(cls, "__annotations__", {})
        if chunkable:
            chunkable = {k: v for k, v in chunkable.items() if isinstance(v, int)}
            cls._chunkable = dict(cls._chunkable, **chunkable)
        if detachable:
            cls._detachable = cls._detachable.union(detachable)
        if serialize_ignore:
            cls._serialize_ignore = cls._serialize_ignore.union(serialize_ignore)
        super().__init_subclass__(**kwargs)


# T = TypeVar("T")

# how i wish the code below would be correct, but we're also parsing into floats
# and converting into strings if the original type is string, but the value isn't
# def _validate_type(t: type, value: T) -> Tuple[bool, T]:


def _validate_type(t: Optional[type], value: Any) -> Tuple[bool, Any]:
    # this should be reworked. Its only ok to return null for Optionals...
    # if t is None and value is None:
    if value is None:
        return True, value

    # after fixing the None t above, this should be
    # if t is Any:
    # if t is None:

    if t is None or t is Any:
        return True, value

    if isclass(t) and issubclass(t, Enum):
        if isinstance(value, t):
            return True, value
        if value in t._value2member_map_:
            return True, t(value)

    if getattr(t, "__module__", None) == "typing":
        if isinstance(t, ForwardRef):
            return True, value

        origin = getattr(t, "__origin__")
        # below is what in nicer for >= py38
        # origin = get_origin(t)

        # recursive validation for Unions on both types preferring the fist type
        if origin is Union:
            # below is what in nicer for >= py38
            # t_1, t_2 = get_args(t)
            args = t.__args__  # type: ignore
            for arg_t in args:
                t_success, t_value = _validate_type(arg_t, value)
                if t_success:
                    return True, t_value
            return False, value
        if origin is dict:
            if not isinstance(value, dict):
                return False, value
            if value == {}:
                return True, value
            if not getattr(t, "__args__", None):
                return True, value
            t_key, t_value = t.__args__  # type: ignore

            if (
                getattr(t_key, "__name__", None),
                getattr(t_value, "__name__", None),
            ) == ("KT", "VT"):
                return True, value
            # we're only checking the first item, but the for loop and return after
            # evaluating the first item is the fastest way
            for dict_key, dict_value in value.items():
                valid_key, _ = _validate_type(t_key, dict_key)
                valid_value, _ = _validate_type(t_value, dict_value)

                if valid_key and valid_value:
                    return True, value
                return False, value

        if origin is list:
            if not isinstance(value, list):
                return False, value
            if value == []:
                return True, value
            if not hasattr(t, "__args__"):
                return True, value
            t_items = t.__args__[0]  # type: ignore
            if getattr(t_items, "__name__", None) == "T":
                return True, value
            first_item_valid, _ = _validate_type(t_items, value[0])
            if first_item_valid:
                return True, value
            return False, value

        if origin is tuple:
            if not isinstance(value, tuple):
                return False, value
            if not hasattr(t, "__args__"):
                return True, value
            args = t.__args__  # type: ignore
            if args == tuple():
                return True, value
            # we're not checking for empty tuple, cause tuple lengths must match
            if len(args) != len(value):
                return False, value
            values = []
            for t_item, v_item in zip(args, value):
                item_valid, item_value = _validate_type(t_item, v_item)
                if not item_valid:
                    return False, value
                values.append(item_value)
            return True, tuple(values)

        if origin is set:
            if not isinstance(value, set):
                return False, value
            if not hasattr(t, "__args__"):
                return True, value
            t_items = t.__args__[0]  # type: ignore
            first_item_valid, _ = _validate_type(t_items, next(iter(value)))
            if first_item_valid:
                return True, value
            return False, value

    if isinstance(value, t):
        return True, value

    with contextlib.suppress(ValueError, TypeError):
        if t is float and value is not None:
            return True, float(value)
        # TODO: dafuq, i had to add this not list check
        # but it would also fail for objects and other complex values
        if t is str and value and not isinstance(value, list):
            return True, str(value)

    return False, value


class Base(_RegisteringBase):
    id: Union[str, None] = None
    totalChildrenCount: Union[int, None] = None
    applicationId: Union[str, None] = None
    _units: Union[None, str] = None

    def __init__(self, **kwargs) -> None:
        super().__init__()
        for k, v in kwargs.items():
            self.__setattr__(k, v)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id: {self.id}, "
            f"speckle_type: {self.speckle_type}, "
            f"totalChildrenCount: {self.totalChildrenCount})"
        )

    def __str__(self) -> str:
        return self.__repr__()

    @classmethod
    def of_type(cls, speckle_type: str, **kwargs) -> "Base":
        """
        Get a plain Base object with a specified speckle_type.

        The speckle_type is protected and cannot be overwritten on a class instance.
        This is to prevent problems with receiving in other platforms or connectors.
        However, if you really need a base with a different type, here is a helper
        to do that for you.

        This is used in the deserialisation of unknown types so their speckle_type
        can be preserved.
        """
        b = cls(**kwargs)
        b.__dict__.update(speckle_type=speckle_type)
        return b

    def __setitem__(self, name: str, value: Any) -> None:
        self.validate_prop_name(name)
        self.__dict__[name] = value

    def __getitem__(self, name: str) -> Any:
        return self.__dict__[name]

    def __setattr__(self, name: str, value: Any) -> None:
        """
        Type checking, guard attribute, and property set mechanism.

        The `speckle_type` is a protected class attribute it must not be overridden.

        This also performs a type check if the attribute is type hinted.
        """
        if name == "speckle_type":
            # not sure if we should raise an exception here??
            # raise SpeckleException(
            #     "Cannot override the `speckle_type`. This is set manually by the class or on deserialisation"
            # )
            return
        # if value is not None:
        value = self._type_check(name, value)
        attr = getattr(self.__class__, name, None)
        if isinstance(attr, property):
            try:
                attr.__set__(self, value)
            except AttributeError:
                return  # the prop probably doesn't have a setter
        super().__setattr__(name, value)

    @classmethod
    def update_forward_refs(cls) -> None:
        """
        Attempts to populate the internal defined types dict for type checking
        sometime after defining the class.
        This is already done when defining the class, but can be called
        again if references to undefined types were
        included.

        See `objects.geometry` for an example of how this is used with
        the Brep class definitions.
        """
        try:
            cls._attr_types = get_type_hints(cls)
        except Exception as e:
            warn(f"Could not update forward refs for class {cls.__name__}: {e}")

    @classmethod
    def validate_prop_name(cls, name: str) -> None:
        """Validator for dynamic attribute names."""
        if name in {"", "@"}:
            raise ValueError("Invalid Name: Base member names cannot be empty strings")
        if name.startswith("@@"):
            raise ValueError(
                "Invalid Name: Base member names cannot start with more than one '@'",
            )
        if "." in name or "/" in name:
            raise ValueError(
                "Invalid Name: Base member names cannot contain characters '.' or '/'",
            )

    def _type_check(self, name: str, value: Any) -> Any:
        """
        Lightweight type checking of values before setting them

        NOTE: Does not check subscripted types within generics as the performance hit
        of checking each item within a given collection isn't worth it.
        Eg if you have a type Dict[str, float],
        we will only check if the value you're trying to set is a dict.
        """
        types = getattr(self, "_attr_types", {})
        t = types.get(name, None)

        valid, checked_value = _validate_type(t, value)

        if valid:
            return checked_value

        raise SpeckleException(
            f"Cannot set '{self.__class__.__name__}.{name}':"
            f"it expects type '{str(t)}',"
            f"but received type '{type(value).__name__}'"
        )

    def add_chunkable_attrs(self, **kwargs: int) -> None:
        """
        Mark defined attributes as chunkable for serialisation

        Arguments:
            kwargs {int} -- the name of the attribute as the keyword
            and the chunk size as the arg
        """
        chunkable = {k: v for k, v in kwargs.items() if isinstance(v, int)}
        self._chunkable = dict(self._chunkable, **chunkable)

    def add_detachable_attrs(self, names: Set[str]) -> None:
        """
        Mark defined attributes as detachable for serialisation

        Arguments:
            names {Set[str]} -- the names of the attributes to detach as a set of string
        """
        self._detachable = self._detachable.union(names)

    @property
    def units(self) -> Union[str, None]:
        return self._units

    @units.setter
    def units(self, value: Union[str, Units, None]):
        """While this property accepts any string value, geometry expects units to be specific strings (see Units enum)"""
        if isinstance(value, str) or value is None:
            self._units = value
        elif isinstance(value, Units):
            self._units = value.value
        else:
            raise SpeckleInvalidUnitException(
                f"Unknown type {type(value)} received for units"
            )

    def get_member_names(self) -> List[str]:
        """Get all of the property names on this object, dynamic or not"""
        attr_dir = list(set(dir(self)) - REMOVE_FROM_DIR)
        return [
            name
            for name in attr_dir
            if not name.startswith("_") and not callable(getattr(self, name))
        ]

    def get_serializable_attributes(self) -> List[str]:
        """Get the attributes that should be serialized"""
        return sorted(list(set(self.get_member_names()) - self._serialize_ignore))

    def get_typed_member_names(self) -> List[str]:
        """Get all of the names of the defined (typed) properties of this object"""
        return list(self._attr_types.keys())

    def get_dynamic_member_names(self) -> List[str]:
        """Get all of the names of the dynamic properties of this object"""
        return list(set(self.__dict__.keys()) - set(self._attr_types.keys()))

    def get_children_count(self) -> int:
        """Get the total count of children Base objects"""
        parsed = []
        return 1 + self._count_descendants(self, parsed)

    def get_id(self, decompose: bool = False) -> str:
        """
        Gets the id (a unique hash) of this object.
        ⚠️ This method fully serializes the object which,
        in the case of large objects (with many sub-objects), has a tangible cost.
        Avoid using it!

        Note: the hash of a decomposed object differs from that of a
        non-decomposed object

        Arguments:
            decompose {bool} -- if True, will decompose the object in
            the process of hashing it

        Returns:
            str -- the hash (id) of the fully serialized object
        """
        from specklepy.serialization.base_object_serializer import BaseObjectSerializer

        serializer = BaseObjectSerializer()
        if decompose:
            serializer.write_transports = [MemoryTransport()]
        return serializer.traverse_base(self)[0]

    def _count_descendants(self, base: "Base", parsed: List) -> int:
        if base in parsed:
            return 0
        parsed.append(base)

        return sum(
            self._handle_object_count(value, parsed)
            for name, value in base.get_member_names()
            if not name.startswith("@")
        )

    def _handle_object_count(self, obj: Any, parsed: List) -> int:
        # pylint: disable=isinstance-second-argument-not-valid-type
        count = 0
        if obj is None:
            return count
        if isinstance(obj, "Base"):
            count += 1
            count += self._count_descendants(obj, parsed)
            return count
        elif isinstance(obj, list):
            for item in obj:
                if isinstance(item, "Base"):
                    count += 1
                    count += self._count_descendants(item, parsed)
                else:
                    count += self._handle_object_count(item, parsed)
        elif isinstance(obj, dict):
            for _, value in obj.items():
                if isinstance(value, "Base"):
                    count += 1
                    count += self._count_descendants(value, parsed)
                else:
                    count += self._handle_object_count(value, parsed)
        return count


Base.update_forward_refs()


class DataChunk(Base, speckle_type="Speckle.Core.Models.DataChunk"):
    data: Union[List[Any], None] = None

    def __init__(self) -> None:
        super().__init__()
        self.data = []



from enum import Enum
from typing import Optional

from specklepy.objects.base import Base

STRUCTURAL_MATERIALS = "Objects.Structural.Materials"


class MaterialType(int, Enum):
    Concrete = 0
    Steel = 1
    Timber = 2
    Aluminium = 3
    Masonry = 4
    FRP = 5
    Glass = 6
    Fabric = 7
    Rebar = 8
    Tendon = 9
    ColdFormed = 10
    Other = 11


class StructuralMaterial(
    Base, speckle_type=STRUCTURAL_MATERIALS + ".StructuralMaterial"
):
    name: Optional[str] = None
    grade: Optional[str] = None
    materialType: Optional[MaterialType] = None
    designCode: Optional[str] = None
    codeYear: Optional[str] = None
    strength: float = 0.0
    elasticModulus: float = 0.0
    poissonsRatio: float = 0.0
    shearModulus: float = 0.0
    density: float = 0.0
    thermalExpansivity: float = 0.0
    dampingRatio: float = 0.0
    cost: float = 0.0
    materialSafetyFactor: float = 0.0


class Concrete(StructuralMaterial):
    compressiveStrength: float = 0.0
    tensileStrength: float = 0.0
    flexuralStrength: float = 0.0
    maxCompressiveStrain: float = 0.0
    maxTensileStrain: float = 0.0
    maxAggregateSize: float = 0.0
    lightweight: Optional[bool] = None


class Steel(StructuralMaterial, speckle_type=STRUCTURAL_MATERIALS + ".Steel"):
    yieldStrength: float = 0.0
    ultimateStrength: float = 0.0
    maxStrain: float = 0.0
    strainHardeningModulus: float = 0.0


class Timber(StructuralMaterial, speckle_type=STRUCTURAL_MATERIALS + ".Timber"):
    species: Optional[str] = None



from enum import Enum
from typing import List, Optional

from specklepy.objects.base import Base
from specklepy.objects.geometry import Line, Mesh, Plane, Point, Vector
from specklepy.objects.structural.axis import Axis
from specklepy.objects.structural.properties import (
    Property1D,
    Property2D,
    Property3D,
    PropertyDamper,
    PropertyMass,
    PropertySpring,
)

STRUCTURAL_GEOMETRY = "Objects.Structural.Geometry"


class ElementType1D(int, Enum):
    Beam = 0
    Brace = 1
    Bar = 2
    Column = 3
    Rod = 4
    Spring = 5
    Tie = 6
    Strut = 7
    Link = 8
    Damper = 9
    Cable = 10
    Spacer = 11
    Other = 12
    Null = 13


class ElementType2D(int, Enum):
    Quad4 = 0
    Quad8 = 1
    Triangle3 = 2
    Triangle6 = 3


class ElementType3D(int, Enum):
    Brick8 = 0
    Wedge6 = 1
    Pyramid5 = 2
    Tetra4 = 3


class Restraint(Base, speckle_type=STRUCTURAL_GEOMETRY + ".Restraint"):
    code: Optional[str] = None
    stiffnessX: float = 0.0
    stiffnessY: float = 0.0
    stiffnessZ: float = 0.0
    stiffnessXX: float = 0.0
    stiffnessYY: float = 0.0
    stiffnessZZ: float = 0.0


class Node(Base, speckle_type=STRUCTURAL_GEOMETRY + ".Node"):
    name: Optional[str] = None
    basePoint: Optional[Point] = None
    constraintAxis: Optional[Axis] = None
    restraint: Optional[Restraint] = None
    springProperty: Optional[PropertySpring] = None
    massProperty: Optional[PropertyMass] = None
    damperProperty: Optional[PropertyDamper] = None


class Element1D(Base, speckle_type=STRUCTURAL_GEOMETRY + ".Element1D"):
    name: Optional[str] = None
    baseLine: Optional[Line] = None
    property: Optional[Property1D] = None
    type: Optional[ElementType1D] = None
    end1Releases: Optional[Restraint] = None
    end2Releases: Optional[Restraint] = None
    end1Offset: Optional[Vector] = None
    end2Offset: Optional[Vector] = None
    orientationNode: Optional[Node] = None
    orinetationAngle: float = 0.0
    localAxis: Optional[Plane] = None
    parent: Optional[Base] = None
    end1Node: Optional[Node] = None
    end2Node: Optional[Node] = None
    topology: Optional[List] = None
    displayMesh: Optional[Mesh] = None


class Element2D(Base, speckle_type=STRUCTURAL_GEOMETRY + ".Element2D"):
    name: Optional[str] = None
    property: Optional[Property2D] = None
    type: Optional[ElementType2D] = None
    offset: float = 0.0
    orientationAngle: float = 0.0
    parent: Optional[Base] = None
    topology: Optional[List] = None
    displayMesh: Optional[Mesh] = None


class Element3D(Base, speckle_type=STRUCTURAL_GEOMETRY + ".Element3D"):
    name: Optional[str] = None
    baseMesh: Optional[Mesh] = None
    property: Optional[Property3D] = None
    type: Optional[ElementType3D] = None
    orientationAngle: float = 0.0
    parent: Optional[Base] = None
    topology: List


# class Storey needs ependency on built elements first



from typing import List, Optional

from specklepy.objects.base import Base
from specklepy.objects.structural.analysis import Model
from specklepy.objects.structural.geometry import Element1D, Element2D, Element3D, Node

STRUCTURAL_RESULTS = "Objects.Structural.Results."


class Result(Base, speckle_type=STRUCTURAL_RESULTS + "Result"):
    resultCase: Optional[Base] = None
    permutation: Optional[str] = None
    description: Optional[str] = None


class ResultSet1D(Result, speckle_type=STRUCTURAL_RESULTS + "ResultSet1D"):
    results1D: List


class Result1D(Result, speckle_type=STRUCTURAL_RESULTS + "Result1D"):
    element: Optional[Element1D] = None
    position: Optional[float] = None
    dispX: Optional[float] = None
    dispY: Optional[float] = None
    dispZ: Optional[float] = None
    rotXX: Optional[float] = None
    rotYY: Optional[float] = None
    rotZZ: Optional[float] = None
    forceX: Optional[float] = None
    forceY: Optional[float] = None
    forceZ: Optional[float] = None
    momentXX: Optional[float] = None
    momentYY: Optional[float] = None
    momentZZ: Optional[float] = None
    axialStress: Optional[float] = None
    shearStressY: Optional[float] = None
    shearStressZ: Optional[float] = None
    bendingStressYPos: Optional[float] = None
    bendingStressYNeg: Optional[float] = None
    bendingStressZPos: Optional[float] = None
    bendingStressZNeg: Optional[float] = None
    combinedStressMax: Optional[float] = None
    combinedStressMin: Optional[float] = None


class ResultSet2D(Result, speckle_type=STRUCTURAL_RESULTS + "ResultSet2D"):
    results2D: List


class Result2D(Result, speckle_type=STRUCTURAL_RESULTS + "Result2D"):
    element: Optional[Element2D] = None
    position: List
    dispX: Optional[float] = None
    dispY: Optional[float] = None
    dispZ: Optional[float] = None
    forceXX: Optional[float] = None
    forceYY: Optional[float] = None
    forceXY: Optional[float] = None
    momentXX: Optional[float] = None
    momentYY: Optional[float] = None
    momentXY: Optional[float] = None
    shearX: Optional[float] = None
    shearY: Optional[float] = None
    stressTopXX: Optional[float] = None
    stressTopYY: Optional[float] = None
    stressTopZZ: Optional[float] = None
    stressTopXY: Optional[float] = None
    stressTopYZ: Optional[float] = None
    stressTopZX: Optional[float] = None
    stressMidXX: Optional[float] = None
    stressMidYY: Optional[float] = None
    stressMidZZ: Optional[float] = None
    stressMidXY: Optional[float] = None
    stressMidYZ: Optional[float] = None
    stressMidZX: Optional[float] = None
    stressBotXX: Optional[float] = None
    stressBotYY: Optional[float] = None
    stressBotZZ: Optional[float] = None
    stressBotXY: Optional[float] = None
    stressBotYZ: Optional[float] = None
    stressBotZX: Optional[float] = None


class ResultSet3D(Result, speckle_type=STRUCTURAL_RESULTS + "ResultSet3D"):
    results3D: List


class Result3D(Result, speckle_type=STRUCTURAL_RESULTS + "Result3D"):
    element: Optional[Element3D] = None
    position: List
    dispX: Optional[float] = None
    dispY: Optional[float] = None
    dispZ: Optional[float] = None
    stressXX: Optional[float] = None
    stressYY: Optional[float] = None
    stressZZ: Optional[float] = None
    stressXY: Optional[float] = None
    stressYZ: Optional[float] = None
    stressZX: Optional[float] = None


class ResultGlobal(Result, speckle_type=STRUCTURAL_RESULTS + "ResultGlobal"):
    model: Optional[Model] = None
    loadX: Optional[float] = None
    loadY: Optional[float] = None
    loadZ: Optional[float] = None
    loadXX: Optional[float] = None
    loadYY: Optional[float] = None
    loadZZ: Optional[float] = None
    reactionX: Optional[float] = None
    reactionY: Optional[float] = None
    reactionZ: Optional[float] = None
    reactionXX: Optional[float] = None
    reactionYY: Optional[float] = None
    reactionZZ: Optional[float] = None
    mode: Optional[float] = None
    frequency: Optional[float] = None
    loadFactor: Optional[float] = None
    modalStiffness: Optional[float] = None
    modalGeoStiffness: Optional[float] = None
    effMassX: Optional[float] = None
    effMassY: Optional[float] = None
    effMassZ: Optional[float] = None
    effMassXX: Optional[float] = None
    effMassYY: Optional[float] = None
    effMassZZ: Optional[float] = None


class ResultSetNode(Result, speckle_type=STRUCTURAL_RESULTS + "ResultSetNode"):
    resultsNode: List


class ResultNode(Result, speckle_type=STRUCTURAL_RESULTS + " ResultNode"):
    node: Optional[Node] = None
    dispX: Optional[float] = None
    dispY: Optional[float] = None
    dispZ: Optional[float] = None
    rotXX: Optional[float] = None
    rotYY: Optional[float] = None
    rotZZ: Optional[float] = None
    reactionX: Optional[float] = None
    reactionY: Optional[float] = None
    reactionZ: Optional[float] = None
    reactionXX: Optional[float] = None
    reactionYY: Optional[float] = None
    reactionZZ: Optional[float] = None
    constraintX: Optional[float] = None
    constraintY: Optional[float] = None
    constraintZ: Optional[float] = None
    constraintXX: Optional[float] = None
    constraintYY: Optional[float] = None
    constraintZZ: Optional[float] = None
    velX: Optional[float] = None
    velY: Optional[float] = None
    velZ: Optional[float] = None
    velXX: Optional[float] = None
    velYY: Optional[float] = None
    velZZ: Optional[float] = None
    accX: Optional[float] = None
    accY: Optional[float] = None
    accZ: Optional[float] = None
    accXX: Optional[float] = None
    accYY: Optional[float] = None
    accZZ: Optional[float] = None


class ResultSetAll(Base, speckle_type=None):
    resultSet1D: Optional[ResultSet1D] = None
    resultSet2D: Optional[ResultSet2D] = None
    resultSet3D: Optional[ResultSet3D] = None
    resultsGlobal: Optional[ResultGlobal] = None
    resultsNode: Optional[ResultSetNode] = None



import pkgutil
import sys
from importlib import import_module

for _, name, _ in pkgutil.iter_modules(__path__):
    imported_module = import_module("." + name, package=__name__)

    if hasattr(imported_module, "Resource"):
        setattr(sys.modules[__name__], name, imported_module)



from typing import List, Optional, Union

from gql import gql

from specklepy.api.models import Commit
from specklepy.api.resource import ResourceBase
from specklepy.logging import metrics

from specklepy.core.api.resources.commit import Resource as CoreResource
from specklepy.logging.exceptions import SpeckleException


class Resource(CoreResource):
    """API Access class for commits"""

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
        )
        self.schema = Commit

    def get(self, stream_id: str, commit_id: str) -> Commit:
        """
        Gets a commit given a stream and the commit id

        Arguments:
            stream_id {str} -- the stream where we can find the commit
            commit_id {str} -- the id of the commit you want to get

        Returns:
            Commit -- the retrieved commit object
        """
        metrics.track(metrics.SDK, self.account, {"name": "Commit Get"})
        return super().get(stream_id, commit_id)

    def list(self, stream_id: str, limit: int = 10) -> List[Commit]:
        """
        Get a list of commits on a given stream

        Arguments:
            stream_id {str} -- the stream where the commits are
            limit {int} -- the maximum number of commits to fetch (default = 10)

        Returns:
            List[Commit] -- a list of the most recent commit objects
        """
        metrics.track(metrics.SDK, self.account, {"name": "Commit List"})
        return super().list(stream_id, limit)

    def create(
        self,
        stream_id: str,
        object_id: str,
        branch_name: str = "main",
        message: str = "",
        source_application: str = "python",
        parents: Optional[List[str]] = None,
    ) -> Union[str, SpeckleException]:
        """
        Creates a commit on a branch

        Arguments:
            stream_id {str} -- the stream you want to commit to
            object_id {str} -- the hash of your commit object
            branch_name {str}
                -- the name of the branch to commit to (defaults to "main")
            message {str}
                -- optional: a message to give more information about the commit
            source_application{str}
                -- optional: the application from which the commit was created
                (defaults to "python")
            parents {List[str]} -- optional: the id of the parent commits

        Returns:
            str -- the id of the created commit
        """
        metrics.track(metrics.SDK, self.account, {"name": "Commit Create"})
        return super().create(
            stream_id, object_id, branch_name, message, source_application, parents
        )

    def update(self, stream_id: str, commit_id: str, message: str) -> bool:
        """
        Update a commit

        Arguments:
            stream_id {str}
                -- the id of the stream that contains the commit you'd like to update
            commit_id {str} -- the id of the commit you'd like to update
            message {str} -- the updated commit message

        Returns:
            bool -- True if the operation succeeded
        """
        metrics.track(metrics.SDK, self.account, {"name": "Commit Update"})
        return super().update(stream_id, commit_id, message)

    def delete(self, stream_id: str, commit_id: str) -> bool:
        """
        Delete a commit

        Arguments:
            stream_id {str}
                -- the id of the stream that contains the commit you'd like to delete
            commit_id {str} -- the id of the commit you'd like to delete

        Returns:
            bool -- True if the operation succeeded
        """
        metrics.track(metrics.SDK, self.account, {"name": "Commit Delete"})
        return super().delete(stream_id, commit_id)

    def received(
        self,
        stream_id: str,
        commit_id: str,
        source_application: str = "python",
        message: Optional[str] = None,
    ) -> bool:
        """
        Mark a commit object a received by the source application.
        """
        metrics.track(metrics.SDK, self.account, {"name": "Commit Received"})
        return super().received(stream_id, commit_id, source_application, message)



from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Type

from specklepy.logging.exceptions import SpeckleException
from specklepy.objects.base import Base


class CurveTypeEncoding(int, Enum):
    Arc = 0
    Circle = 1
    Curve = 2
    Ellipse = 3
    Line = 4
    Polyline = 5
    Polycurve = 6

    @property
    def object_class(self) -> Type:
        from . import geometry

        if self == self.Arc:
            return geometry.Arc
        elif self == self.Circle:
            return geometry.Circle
        elif self == self.Curve:
            return geometry.Curve
        elif self == self.Ellipse:
            return geometry.Ellipse
        elif self == self.Line:
            return geometry.Line
        elif self == self.Polyline:
            return geometry.Polyline
        elif self == self.Polycurve:
            return geometry.Polycurve
        raise SpeckleException(
            f"No corresponding object class for CurveTypeEncoding: {self}"
        )


def curve_from_list(args: List[float]):
    curve_type = CurveTypeEncoding(args[0])
    return curve_type.object_class.from_list(args)


class ObjectArray:
    def __init__(self, data: Optional[list] = None) -> None:
        self.data = data or []

    @classmethod
    def from_objects(cls, objects: List[Base]) -> "ObjectArray":
        data_list = cls()
        if not objects:
            return data_list

        speckle_type = objects[0].speckle_type

        for obj in objects:
            if speckle_type != obj.speckle_type:
                raise SpeckleException(
                    "All objects in chunk should have the same speckle_type. "
                    f"Found {speckle_type} and {obj.speckle_type}"
                )
            data_list.encode_object(obj=obj)

        return data_list

    @staticmethod
    def decode_data(
        data: List[Any], decoder: Callable[[List[Any]], Base], **kwargs: Dict[str, Any]
    ) -> List[Base]:
        bases: List[Base] = []
        if not data:
            return bases
        index = 0
        while index < len(data):
            item_length = int(data[index])
            item_start = index + 1
            item_end = item_start + item_length
            item_data = data[item_start:item_end]
            index = item_end
            decoded_data = decoder(item_data, **kwargs)
            bases.append(decoded_data)

        return bases

    def decode(self, decoder: Callable[[List[Any]], Any], **kwargs: Dict[str, Any]):
        return self.decode_data(data=self.data, decoder=decoder, **kwargs)

    def encode_object(self, obj: Base):
        encoded = obj.to_list()
        encoded.insert(0, len(encoded))
        self.data.extend(encoded)


class CurveArray(ObjectArray):
    @classmethod
    def from_curve(cls, curve: Base) -> "CurveArray":
        crv_array = cls()
        crv_array.data = curve.to_list()
        return crv_array

    @classmethod
    def from_curves(cls, curves: List[Base]) -> "CurveArray":
        data = []
        for curve in curves:
            curve_list = curve.to_list()
            curve_list.insert(0, len(curve_list))
            data.extend(curve_list)
        crv_array = cls()
        crv_array.data = data
        return crv_array

    @staticmethod
    def curve_from_list(args: List[float]) -> Base:
        curve_type = CurveTypeEncoding(args[0])
        return curve_type.object_class.from_list(args)

    @property
    def type(self) -> CurveTypeEncoding:
        return CurveTypeEncoding(self.data[0])

    def to_curve(self) -> Base:
        return self.type.object_class.from_list(self.data)

    @classmethod
    def _curve_decoder(cls, data: List[float]) -> Base:
        crv_array = cls(data)
        return crv_array.to_curve()

    def to_curves(self) -> List[Base]:
        return self.decode(decoder=self._curve_decoder)



"""Builtin Speckle object kit."""

from specklepy.objects.GIS.layers import (
    VectorLayer,
    RasterLayer,
)

from specklepy.objects.GIS.geometry import (
    GisPolygonGeometry,
    GisPolygonElement,
    GisLineElement,
    GisPointElement,
    GisRasterElement,
)

from specklepy.objects.GIS.CRS import (
    CRS,
)

__all__ = ["VectorLayer", "RasterLayer", 
           "GisPolygonGeometry", "GisPolygonElement", "GisLineElement", "GisPointElement", "GisRasterElement", 
           "CRS"]
